__license__   = 'GPL v3'
__copyright__ = '2008, Kovid Goyal <kovid at kovidgoyal.net>'

from qt.core import QAbstractItemView, QDialog, QSortFilterProxyModel, QStringListModel, Qt

from calibre.constants import islinux
from calibre.gui2 import error_dialog, gprefs, question_dialog
from calibre.gui2.dialogs.confirm_delete import confirm
from calibre.gui2.dialogs.tag_editor_ui import Ui_TagEditor
from calibre.startup import connect_lambda
from calibre.utils.icu import sort_key


class TagEditor(QDialog, Ui_TagEditor):

    def __init__(self, window, db, id_=None, key=None, current_tags=None):
        QDialog.__init__(self, window)
        Ui_TagEditor.__init__(self)
        self.setupUi(self)

        self.db = db
        self.sep = ','
        self.is_names = False
        if key:
            # Assume that if given a key then it is a custom column
            try:
                fm = db.field_metadata[key]
                self.is_names = fm['display'].get('is_names', False)
                if self.is_names:
                    self.sep = '&'
                self.setWindowTitle(self.windowTitle() + ': ' + fm['name'])
            except Exception:
                pass
            key = db.field_metadata.key_to_label(key)
        else:
            self.setWindowTitle(self.windowTitle() + ': ' + db.field_metadata['tags']['name'])

        if self.sep == '&':
            self.add_tag_input.setToolTip('<p>' +
                        _('If the item you want is not in the available list, '
                          'you can add it here. Accepts an ampersand-separated '
                          'list of items. The items will be applied to '
                          'the book.') + '</p>')
        else:
            self.add_tag_input.setToolTip('<p>' +
                        _('If the item you want is not in the available list, '
                          'you can add it here. Accepts a comma-separated '
                          'list of items. The items will be applied to '
                          'the book.') + '</p>')
        self.key = key
        self.index = db.row(id_) if id_ is not None else None
        if self.index is not None:
            if key is None:
                tags = self.db.tags(self.index)
                if tags:
                    tags = [tag.strip() for tag in tags.split(',') if tag.strip()]
            else:
                tags = self.db.get_custom(self.index, label=key)
        else:
            tags = []
        if current_tags is not None:
            tags = sorted(set(current_tags), key=sort_key)
        if tags:
            if not self.is_names:
                tags.sort(key=sort_key)
        else:
            tags = []
        self.applied_model = QStringListModel(tags)
        p = QSortFilterProxyModel()
        p.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        p.setSourceModel(self.applied_model)
        self.applied_tags.setModel(p)
        if self.is_names:
            self.applied_tags.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
            self.applied_tags.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        if key:
            all_tags = [tag for tag in self.db.all_custom(label=key)]
        else:
            all_tags = [tag for tag in self.db.all_tags()]
        all_tags = sorted(set(all_tags) - set(tags), key=sort_key)
        self.all_tags_model = QStringListModel(all_tags)
        p = QSortFilterProxyModel()
        p.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        p.setSourceModel(self.all_tags_model)
        self.available_tags.setModel(p)

        connect_lambda(self.apply_button.clicked, self, lambda self: self.apply_tags())
        connect_lambda(self.unapply_button.clicked, self, lambda self: self.unapply_tags())
        self.add_tag_button.clicked.connect(self.add_tag)
        connect_lambda(self.delete_button.clicked, self, lambda self: self.delete_tags())
        self.add_tag_input.returnPressed[()].connect(self.add_tag)
        # add the handlers for the filter input fields
        connect_lambda(self.available_filter_input.textChanged, self, lambda self, text: self.filter_tags(text))
        connect_lambda(self.applied_filter_input.textChanged, self, lambda self, text: self.filter_tags(text, which='applied_tags'))

        # Restore the focus to the last input box used (typed into)
        for x in ('add_tag_input', 'available_filter_input', 'applied_filter_input'):
            ibox = getattr(self, x)
            ibox.setObjectName(x)
            connect_lambda(ibox.textChanged, self, lambda self: self.edit_box_changed(self.sender().objectName()))
        getattr(self, gprefs.get('tag_editor_last_filter', 'add_tag_input')).setFocus()

        self.available_tags.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.applied_tags.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        if islinux:
            self.available_tags.doubleClicked.connect(self.apply_tags)
            self.applied_tags.doubleClicked.connect(self.unapply_tags)
        else:
            self.available_tags.activated.connect(self.apply_tags)
            self.applied_tags.activated.connect(self.unapply_tags)

        self.restore_geometry(gprefs, 'tag_editor_geometry')

    def edit_box_changed(self, which):
        gprefs['tag_editor_last_filter'] = which

    def delete_tags(self):
        confirms, deletes = [], []
        row_indices = list(self.available_tags.selectionModel().selectedRows())

        if not row_indices:
            error_dialog(self, _('No tags selected'), _('You must select at least one tag from the list of Available tags.')).exec()
            return
        if not confirm(
            _('Deleting tags is done immediately and there is no undo.'),
            'tag_editor_delete'):
            return
        pos = self.available_tags.verticalScrollBar().value()
        for ri in row_indices:
            tag = ri.data()
            used = self.db.is_tag_used(tag) \
                if self.key is None else \
                self.db.is_item_used_in_multiple(tag, label=self.key)
            if used:
                confirms.append(ri)
            else:
                deletes.append(ri)
        if confirms:
            ct = ', '.join(item.data() for item in confirms)
            if question_dialog(self, _('Are your sure?'),
                '<p>'+_('The following tags are used by one or more books. '
                    'Are you certain you want to delete them?')+'<br>'+ct):
                deletes += confirms

        for item in sorted(deletes, key=lambda r: r.row(), reverse=True):
            tag = item.data()
            if self.key is None:
                self.db.delete_tag(tag)
            else:
                bks = self.db.delete_item_from_multiple(tag, label=self.key)
                self.db.refresh_ids(bks)
            self.available_tags.model().removeRows(item.row(), 1)
        self.available_tags.verticalScrollBar().setValue(pos)

    def apply_tags(self, item=None):
        row_indices = list(self.available_tags.selectionModel().selectedRows())
        row_indices.sort(key=lambda r: r.row(), reverse=True)
        if not row_indices:
            text = self.available_filter_input.text()
            if text and text.strip():
                self.add_tag_input.setText(text)
                self.add_tag_input.setFocus(Qt.FocusReason.OtherFocusReason)
            return
        pos = self.available_tags.verticalScrollBar().value()
        tags = self._get_applied_tags_box_contents()
        for item in row_indices:
            tag = item.data()
            tags.append(tag)
            self.available_tags.model().removeRows(item.row(), 1)
        self.available_tags.verticalScrollBar().setValue(pos)

        if not self.is_names:
            tags.sort(key=sort_key)
        self.applied_model.setStringList(tags)

    def _get_applied_tags_box_contents(self):
        return list(self.applied_model.stringList())

    def unapply_tags(self, item=None):
        row_indices = list(self.applied_tags.selectionModel().selectedRows())
        tags = [r.data() for r in row_indices]
        row_indices.sort(key=lambda r: r.row(), reverse=True)
        for item in row_indices:
            self.applied_model.removeRows(item.row(), 1)

        all_tags = self.all_tags_model.stringList() + tags
        all_tags.sort(key=sort_key)
        self.all_tags_model.setStringList(all_tags)

    def add_tag(self):
        tags = str(self.add_tag_input.text()).split(self.sep)
        tags_in_box = self._get_applied_tags_box_contents()
        for tag in tags:
            tag = tag.strip()
            if not tag:
                continue
            if self.all_tags_model.rowCount():
                for index in self.all_tags_model.match(self.all_tags_model.index(0), Qt.ItemDataRole.DisplayRole, tag, -1,
                                                    Qt.MatchFlag.MatchFixedString | Qt.MatchFlag.MatchCaseSensitive | Qt.MatchFlag.MatchWrap):
                    self.all_tags_model.removeRow(index.row())
            if tag not in tags_in_box:
                tags_in_box.append(tag)

        if not self.is_names:
            tags_in_box.sort(key=sort_key)
        self.applied_model.setStringList(tags_in_box)
        self.add_tag_input.setText('')

    def filter_tags(self, filter_value, which='available_tags'):
        collection = getattr(self, which)
        collection.model().setFilterFixedString(filter_value or '')

    def accept(self):
        if self.add_tag_input.text().strip():
            self.add_tag()
        self.tags = self._get_applied_tags_box_contents()
        self.save_state()
        return QDialog.accept(self)

    def reject(self):
        self.save_state()
        return QDialog.reject(self)

    def save_state(self):
        self.save_geometry(gprefs, 'tag_editor_geometry')


if __name__ == '__main__':
    from calibre.gui2 import Application
    from calibre.library import db
    db = db()
    app = Application([])
    d = TagEditor(None, db, current_tags='a b c'.split())

    if d.exec() == QDialog.DialogCode.Accepted:
        print(d.tags)



#!/usr/bin/env python


__license__   = 'GPL v3'
__copyright__ = '2013, Kovid Goyal <kovid at kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import sys
import traceback
import weakref
from threading import Event, Thread

from calibre.ebooks.metadata.opf2 import metadata_to_opf


def prints(*a, **kw):
    kw['file'] = sys.stderr
    return print(*a, **kw)


class Abort(Exception):
    pass


class MetadataBackup(Thread):
    '''
    Continuously backup changed metadata into OPF files
    in the book directory. This class runs in its own
    thread.
    '''

    def __init__(self, db, interval=2, scheduling_interval=0.1):
        Thread.__init__(self)
        self.daemon = True
        self._db = weakref.ref(getattr(db, 'new_api', db))
        self.stop_running = Event()
        self.interval = interval
        self.scheduling_interval = scheduling_interval
        self.check_dirtied_annotations = 0

    @property
    def db(self):
        ans = self._db()
        if ans is None:
            raise Abort()
        return ans

    def stop(self):
        self.stop_running.set()

    def wait(self, interval):
        if self.stop_running.wait(interval):
            raise Abort()

    def run(self):
        while not self.stop_running.is_set():
            try:
                self.wait(self.interval)
                self.do_one()
            except Abort:
                break

    def do_one(self):
        self.check_dirtied_annotations += 1
        if self.check_dirtied_annotations > 2:
            self.check_dirtied_annotations = 0
            try:
                self.db.check_dirtied_annotations()
            except Exception:
                if self.stop_running.is_set() or self.db.is_closed:
                    return
                traceback.print_exc()

        try:
            book_id = self.db.get_a_dirtied_book()
            if book_id is None:
                return
        except Abort:
            raise
        except:
            # Happens during interpreter shutdown
            return

        self.wait(0)

        try:
            mi, sequence = self.db.get_metadata_for_dump(book_id)
        except:
            prints('Failed to get backup metadata for id:', book_id, 'once')
            traceback.print_exc()
            self.wait(self.interval)
            try:
                mi, sequence = self.db.get_metadata_for_dump(book_id)
            except:
                prints('Failed to get backup metadata for id:', book_id, 'again, giving up')
                traceback.print_exc()
                return

        if mi is None:
            self.db.clear_dirtied(book_id, sequence)
            return

        # Give the GUI thread a chance to do something. Python threads don't
        # have priorities, so this thread would naturally keep the processor
        # until some scheduling event happens. The wait makes such an event
        self.wait(self.scheduling_interval)

        try:
            raw = metadata_to_opf(mi)
        except:
            prints('Failed to convert to opf for id:', book_id)
            traceback.print_exc()
            self.db.clear_dirtied(book_id, sequence)
            return

        self.wait(self.scheduling_interval)

        try:
            self.db.write_backup(book_id, raw)
        except:
            prints('Failed to write backup metadata for id:', book_id, 'once')
            traceback.print_exc()
            self.wait(self.interval)
            try:
                self.db.write_backup(book_id, raw)
            except:
                prints('Failed to write backup metadata for id:', book_id, 'again, giving up')
                traceback.print_exc()
                return

        self.db.clear_dirtied(book_id, sequence)

    def break_cycles(self):
        # Legacy compatibility
        pass



#!/usr/bin/env python

__license__   = 'GPL v3'
__copyright__ = '2008, Kovid Goyal <kovid at kovidgoyal.net>'
'''
Builtin recipes.
'''
import io
import re
import time

from calibre.ebooks.BeautifulSoup import BeautifulSoup
from calibre.utils.config import JSONConfig
from calibre.web.feeds.news import AutomaticNewsRecipe, BasicNewsRecipe, CalibrePeriodical, CustomIndexRecipe
from polyglot.builtins import codepoint_to_chr, itervalues

basic_recipes = (BasicNewsRecipe, AutomaticNewsRecipe, CustomIndexRecipe,
        CalibrePeriodical)

custom_recipes = JSONConfig('custom_recipes/index.json')


def custom_recipe_filename(id_, title):
    from calibre.utils.filenames import ascii_filename
    return ascii_filename(title[:50]) + \
                        ('_%s.recipe'%id_)


def compile_recipe(src):
    '''
    Compile the code in src and return a recipe object, if found.

    :param src: Python source code as bytestring or unicode object

    :return: Recipe class or None, if no such class was found in src
    '''
    if not isinstance(src, str):
        match = re.search(br'coding[:=]\s*([-\w.]+)', src[:200])
        enc = match.group(1).decode('utf-8') if match else 'utf-8'
        src = src.decode(enc)
    # Python complains if there is a coding declaration in a unicode string
    src = re.sub(r'^#.*coding\s*[:=]\s*([-\w.]+)', '#', src.lstrip('\ufeff'), flags=re.MULTILINE)
    # Translate newlines to \n
    src = io.StringIO(src, newline=None).getvalue()

    namespace = {
            'BasicNewsRecipe':BasicNewsRecipe,
            'AutomaticNewsRecipe':AutomaticNewsRecipe,
            'time':time, 're':re,
            'BeautifulSoup':BeautifulSoup,
            'unicode': str,
            'unichr': codepoint_to_chr,
            'xrange': range,
    }
    exec(src, namespace)
    ua = namespace.get('calibre_most_common_ua')

    for x in itervalues(namespace):
        if (isinstance(x, type) and issubclass(x, BasicNewsRecipe) and x not
                in basic_recipes):
            x.calibre_most_common_ua = ua
            return x

    return None



# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/convert/font_key.ui'
#
# Created by: PyQt6 UI code generator 6.4.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(536, 554)
        icon = QtGui.QIcon.ic("wizard.png")
        Dialog.setWindowIcon(icon)
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok|QtWidgets.QDialogButtonBox.StandardButton.RestoreDefaults)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 5, 1, 1, 3)
        self.wh_label = QtWidgets.QLabel(Dialog)
        self.wh_label.setWordWrap(True)
        self.wh_label.setOpenExternalLinks(True)
        self.wh_label.setObjectName("wh_label")
        self.gridLayout.addWidget(self.wh_label, 0, 0, 1, 4)
        self.groupBox_2 = QtWidgets.QGroupBox(Dialog)
        self.groupBox_2.setObjectName("groupBox_2")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox_2)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_3 = QtWidgets.QLabel(self.groupBox_2)
        self.label_3.setObjectName("label_3")
        self.gridLayout_2.addWidget(self.label_3, 0, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.groupBox_2)
        self.label_4.setObjectName("label_4")
        self.gridLayout_2.addWidget(self.label_4, 1, 0, 1, 2)
        self.font_size_key = QtWidgets.QLineEdit(self.groupBox_2)
        self.font_size_key.setClearButtonEnabled(True)
        self.font_size_key.setObjectName("font_size_key")
        self.gridLayout_2.addWidget(self.font_size_key, 1, 2, 1, 1)
        self.output_base_font_size = QtWidgets.QDoubleSpinBox(self.groupBox_2)
        self.output_base_font_size.setDecimals(1)
        self.output_base_font_size.setObjectName("output_base_font_size")
        self.gridLayout_2.addWidget(self.output_base_font_size, 0, 2, 1, 1)
        self.button_use_default = QtWidgets.QPushButton(self.groupBox_2)
        self.button_use_default.setObjectName("button_use_default")
        self.gridLayout_2.addWidget(self.button_use_default, 2, 0, 1, 3)
        self.gridLayout.addWidget(self.groupBox_2, 2, 0, 1, 4)
        self.groupBox = QtWidgets.QGroupBox(Dialog)
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_5 = QtWidgets.QLabel(self.groupBox)
        self.label_5.setObjectName("label_5")
        self.gridLayout_3.addWidget(self.label_5, 0, 0, 1, 1)
        self.input_base_font_size = QtWidgets.QDoubleSpinBox(self.groupBox)
        self.input_base_font_size.setDecimals(1)
        self.input_base_font_size.setObjectName("input_base_font_size")
        self.gridLayout_3.addWidget(self.input_base_font_size, 0, 1, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_6 = QtWidgets.QLabel(self.groupBox)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout.addWidget(self.label_6)
        self.input_font_size = QtWidgets.QDoubleSpinBox(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.input_font_size.sizePolicy().hasHeightForWidth())
        self.input_font_size.setSizePolicy(sizePolicy)
        self.input_font_size.setMaximumSize(QtCore.QSize(130, 16777215))
        self.input_font_size.setDecimals(1)
        self.input_font_size.setObjectName("input_font_size")
        self.horizontalLayout.addWidget(self.input_font_size)
        self.label_7 = QtWidgets.QLabel(self.groupBox)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout.addWidget(self.label_7)
        self.input_mapped_font_size = QtWidgets.QLabel(self.groupBox)
        self.input_mapped_font_size.setObjectName("input_mapped_font_size")
        self.horizontalLayout.addWidget(self.input_mapped_font_size)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.gridLayout_3.addLayout(self.horizontalLayout, 1, 0, 1, 2)
        self.gridLayout.addWidget(self.groupBox, 1, 0, 1, 4)
        self.label_3.setBuddy(self.output_base_font_size)
        self.label_4.setBuddy(self.font_size_key)
        self.label_5.setBuddy(self.input_base_font_size)
        self.label_6.setBuddy(self.input_font_size)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):

        Dialog.setWindowTitle(_("Font rescaling wizard"))
        self.wh_label.setText(_("<p>This wizard will help you choose an appropriate font size key for your needs. Just enter the base font size of the input document and then enter an input font size. The wizard will display what font size it will be mapped to, by the font rescaling algorithm. You can adjust the algorithm by adjusting the output base font size and font key below. When you find values suitable for you, click OK.</p>\n"
"<p>By default, if the output base font size is zero and/or no font size key is specified, calibre will use the values from the current Output profile. </p>\n"
"<p>See the <a href=\"%s\">User Manual</a> for an explanation of how font size rescaling works.</p>"))
        self.groupBox_2.setTitle(_("&Output document"))
        self.label_3.setText(_("&Base font size:"))
        self.label_4.setText(_("Font size &key:"))
        self.output_base_font_size.setSuffix(_(" pt"))
        self.button_use_default.setText(_("Use &default values"))
        self.groupBox.setTitle(_("&Input document"))
        self.label_5.setText(_("&Base font size:"))
        self.input_base_font_size.setSuffix(_(" pt"))
        self.label_6.setText(_("&Font size: "))
        self.input_font_size.setSuffix(_(" pt"))
        self.label_7.setText(_(" will map to size: "))
        self.input_mapped_font_size.setText(_("0.0 pt"))



# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/preferences/template_functions.ui'
#
# Created by: PyQt6 UI code generator 6.4.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(788, 663)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.tabWidget = ScrollingTabWidget(Form)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.gridLayout1 = QtWidgets.QGridLayout(self.tab)
        self.gridLayout1.setContentsMargins(0, 0, 0, 0)
        self.gridLayout1.setObjectName("gridLayout1")
        self.st_show_hide_help_button = QtWidgets.QPushButton(self.tab)
        self.st_show_hide_help_button.setObjectName("st_show_hide_help_button")
        self.gridLayout1.addWidget(self.st_show_hide_help_button, 0, 0, 1, 1)
        self.st_textBrowser = QtWidgets.QTextBrowser(self.tab)
        self.st_textBrowser.setObjectName("st_textBrowser")
        self.gridLayout1.addWidget(self.st_textBrowser, 0, 1, 2, 1)
        self.template_editor = EmbeddedTemplateDialog(self.tab)
        self.template_editor.setObjectName("template_editor")
        self.gridLayout1.addWidget(self.template_editor, 2, 1, 1, 1)
        self.st_button_layout = QtWidgets.QVBoxLayout()
        self.st_button_layout.setObjectName("st_button_layout")
        self.st_clear_button = QtWidgets.QPushButton(self.tab)
        self.st_clear_button.setObjectName("st_clear_button")
        self.st_button_layout.addWidget(self.st_clear_button)
        self.st_delete_button = QtWidgets.QPushButton(self.tab)
        self.st_delete_button.setObjectName("st_delete_button")
        self.st_button_layout.addWidget(self.st_delete_button)
        self.st_replace_button = QtWidgets.QPushButton(self.tab)
        self.st_replace_button.setObjectName("st_replace_button")
        self.st_button_layout.addWidget(self.st_replace_button)
        self.st_create_button = QtWidgets.QPushButton(self.tab)
        self.st_create_button.setObjectName("st_create_button")
        self.st_button_layout.addWidget(self.st_create_button)
        spacerItem = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.st_button_layout.addItem(spacerItem)
        self.st_test_template_button = QtWidgets.QPushButton(self.tab)
        self.st_test_template_button.setObjectName("st_test_template_button")
        self.st_button_layout.addWidget(self.st_test_template_button)
        spacerItem1 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.st_button_layout.addItem(spacerItem1)
        self.gridLayout1.addLayout(self.st_button_layout, 2, 0, 1, 1)
        self.tabWidget.addTab(self.tab, "")
        self.tab1 = QtWidgets.QWidget()
        self.tab1.setObjectName("tab1")
        self.gridLayout2 = QtWidgets.QGridLayout(self.tab1)
        self.gridLayout2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout2.setObjectName("gridLayout2")
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.show_hide_help_button = QtWidgets.QPushButton(self.tab1)
        self.show_hide_help_button.setObjectName("show_hide_help_button")
        self.gridLayout_3.addWidget(self.show_hide_help_button, 0, 0, 1, 1)
        self.textBrowser = QtWidgets.QTextBrowser(self.tab1)
        self.textBrowser.setObjectName("textBrowser")
        self.gridLayout_3.addWidget(self.textBrowser, 0, 1, 2, 2)
        self.show_only_user_defined = QtWidgets.QCheckBox(self.tab1)
        self.show_only_user_defined.setObjectName("show_only_user_defined")
        self.gridLayout_3.addWidget(self.show_only_user_defined, 3, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.tab1)
        self.label_2.setObjectName("label_2")
        self.gridLayout_3.addWidget(self.label_2, 2, 0, 1, 1)
        self.function_name = QtWidgets.QComboBox(self.tab1)
        self.function_name.setEditable(True)
        self.function_name.setObjectName("function_name")
        self.gridLayout_3.addWidget(self.function_name, 2, 1, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.tab1)
        self.label_3.setToolTip("")
        self.label_3.setObjectName("label_3")
        self.gridLayout_3.addWidget(self.label_3, 4, 0, 1, 1)
        self.argument_count = QtWidgets.QSpinBox(self.tab1)
        self.argument_count.setMinimum(-1)
        self.argument_count.setObjectName("argument_count")
        self.gridLayout_3.addWidget(self.argument_count, 4, 1, 1, 1)
        self.documentation = QtWidgets.QTextEdit(self.tab1)
        self.documentation.setObjectName("documentation")
        self.gridLayout_3.addWidget(self.documentation, 5, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.tab1)
        self.label_4.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeading|QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignTop)
        self.label_4.setObjectName("label_4")
        self.gridLayout_3.addWidget(self.label_4, 5, 0, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.clear_button = QtWidgets.QPushButton(self.tab1)
        self.clear_button.setObjectName("clear_button")
        self.horizontalLayout_2.addWidget(self.clear_button)
        self.delete_button = QtWidgets.QPushButton(self.tab1)
        self.delete_button.setObjectName("delete_button")
        self.horizontalLayout_2.addWidget(self.delete_button)
        self.replace_button = QtWidgets.QPushButton(self.tab1)
        self.replace_button.setObjectName("replace_button")
        self.horizontalLayout_2.addWidget(self.replace_button)
        self.create_button = QtWidgets.QPushButton(self.tab1)
        self.create_button.setObjectName("create_button")
        self.horizontalLayout_2.addWidget(self.create_button)
        self.gridLayout_3.addLayout(self.horizontalLayout_2, 6, 1, 1, 1)
        self.label_41 = QtWidgets.QLabel(self.tab1)
        self.label_41.setObjectName("label_41")
        self.gridLayout_3.addWidget(self.label_41, 2, 2, 1, 1)
        self.program = QtWidgets.QPlainTextEdit(self.tab1)
        self.program.setMinimumSize(QtCore.QSize(400, 0))
        self.program.setDocumentTitle("")
        self.program.setTabStopDistance(30)
        self.program.setObjectName("program")
        self.gridLayout_3.addWidget(self.program, 3, 2, 3, 1)
        self.gridLayout2.addLayout(self.gridLayout_3, 1, 0, 1, 1)
        self.tabWidget.addTab(self.tab1, "")
        self.gridLayout.addWidget(self.tabWidget, 0, 0, 1, 1)
        self.label_2.setBuddy(self.function_name)
        self.label_3.setBuddy(self.argument_count)
        self.label_4.setBuddy(self.documentation)
        self.label_41.setBuddy(self.program)

        self.retranslateUi(Form)
        self.tabWidget.setCurrentIndex(0)

    def retranslateUi(self, Form):

        Form.setWindowTitle(_("Form"))
        self.st_show_hide_help_button.setText(_("Hide help"))
        self.st_clear_button.setText(_("&Clear"))
        self.st_delete_button.setText(_("D&elete"))
        self.st_replace_button.setText(_("&Replace"))
        self.st_create_button.setText(_("C&reate"))
        self.st_test_template_button.setText(_("Test"))
        self.st_test_template_button.setToolTip(_("Open a template tester dialog to use a template to test stored templates"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _("&Stored Templates"))
        self.show_hide_help_button.setText(_("Hide help"))
        self.show_only_user_defined.setText(_("Show &only user defined template functions"))
        self.show_only_user_defined.setToolTip(_("<p>If checked the \'Function\' box will show only user defined\n"
"template functions. If unchecked the box will show both system and user defined\n"
"functions. Unchecking can be useful for copying source of a system function to create\n"
"a new related user defined function.</p>"))
        self.label_2.setText(_("F&unction:"))
        self.function_name.setToolTip(_("Enter the name of the function to create."))
        self.label_3.setText(_("Argument &count:"))
        self.argument_count.setToolTip(_("Set this to -1 if the function takes a variable number of arguments"))
        self.label_4.setText(_("D&ocumentation:"))
        self.clear_button.setText(_("Clear"))
        self.delete_button.setText(_("Delete"))
        self.replace_button.setText(_("Replace"))
        self.create_button.setText(_("C&reate"))
        self.label_41.setText(_("P&rogram code (Follow Python indenting rules):"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab1), _("&Template functions"))
from calibre.gui2.dialogs.template_dialog import EmbeddedTemplateDialog
from calibre.gui2.widgets2 import ScrollingTabWidget



#########################################################################
#                                                                       #
#                                                                       #
#   copyright 2002 Paul Henry Tremblay                                  #
#                                                                       #
#   This program is distributed in the hope that it will be useful,     #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of      #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU    #
#   General Public License for more details.                            #
#                                                                       #
#                                                                       #
#########################################################################
import os
import sys

from calibre.ebooks.rtf2xml import copy, field_strings
from calibre.ptempfile import better_mktemp

from . import open_for_read, open_for_write


class FieldsLarge:
    r"""
=========================
Logic
=========================
Make tags for fields.
-Fields reflect text that Microsoft Word automatically generates.
-Each file contains (or should contain) an inner group called field instructions.
-Fields can be nested.
--------------
Logic
--------------
1. As soon as a field is found, make a new text string by appending an empty
text string to the field list. Collect all the lines in this string until the
field instructions are found.
2. Collect all the tokens and text in the field instructions. When the end of
the field instructions is found, process the string of text with the
field_strings module. Append the processed string to the field instructins
list.
3. Continue collecting tokens. Check for paragraphs or sections. If either is found, add to the paragraph or section list.
4. Continue collecting tokens and text either the beginning of a new field is found, or the end of this field is found.
5. If a new field is found, repeat steps 1-3.
6. If the end of the field is found, process the last text string of the field list.
7. If the field list is empty (after removing the last text string), there are
no more fields. Print out the final string. If the list contains other strings,
add the processed string to the last string in the field list.
============================
Examples
============================
    This line of RTF:
        {\field{\*\fldinst { CREATEDATE  \\* MERGEFORMAT }}{\fldrslt {
        \lang1024 1/11/03 10:34 PM}}}
    Becomes:
        <field type = "insert-time">
            10:34 PM
        </field>
    The simple field in the above example contains no paragraph or sections breaks.
    This line of RTF:
        {{\field{\*\fldinst SYMBOL 97 \\f "Symbol" \\s 12}{\fldrslt\f3\fs24}}}
    Becomes:
        <para><inline font-size="18"><inline font-style="Symbol">&#x03A7;</inline></inline></para>
        The RTF in the example above should be represented as UTF-8 rather than a field.
    This RTF:
        {\field\fldedit{\*\fldinst { TOC \\o "1-3" }}{\fldrslt {\lang1024
        Heading one\tab }{\field{\*\fldinst {\lang1024  PAGEREF _Toc440880424
        \\h }{\lang1024 {\*\datafield
        {\lang1024 1}}}{\lang1024 \par }\pard\plain
        \s18\li240\widctlpar\tqr\tldot\tx8630\aspalpha\aspnum\faauto\adjustright\rin0\lin240\itap0
        \f4\lang1033\cgrid {\lang1024 Heading 2\tab }{\field{\*\fldinst
        {\lang1024  PAGEREF _Toc440880425 \\h }{\lang1024 {\*\datafield
        {\lang1024 1}}}{\lang1024 \par }\pard\plain
        \widctlpar\aspalpha\aspnum\faauto\adjustright\rin0\lin0\itap0
        \f4\lang1033\cgrid }}\pard\plain
        \widctlpar\aspalpha\aspnum\faauto\adjustright\rin0\lin0\itap0
        \f4\lang1033\cgrid {\fs28 \\u214\'85 \par }{\fs36 {\field{\*\fldinst
        SYMBOL 67 \\f "Symbol" \\s 18}{\fldrslt\f3\fs36}}}
    Becomes:
        <field-block type="table-of-contents">
        <paragraph-definition language="1033" nest-level="0"
        font-style="Times" name="toc 1" adjust-right="true"
        widow-control="true">
        <para><inline language="1024">Heading one&#x009;</inline><field
        type="reference-to-page" ref="_Toc440880424"><inline
        language="1024">1</inline></field></para>
        </paragraph-definition>
        <paragraph-definition language="1033" nest-level="0" left-indent="12"
        font-style="Times" name="toc 2" adjust-right="true"
        widow-control="true">
        <para><inline language="1024">Heading 2&#x009;</inline><field
        type="reference-to-page" ref="_Toc440880425"><inline
        language="1024">1</inline></field></para>
        </paragraph-definition>
        </field-block>
    """

    def __init__(self,
            in_file,
            bug_handler,
            copy=None,
            run_level=1,
            ):
        """
        Required:
            'file'--file to parse
        Optional:
            'copy'-- whether to make a copy of result for debugging
            'temp_dir' --where to output temporary results (default is
            directory from which the script is run.)
        Returns:
            nothing
            """
        self.__file = in_file
        self.__bug_handler = bug_handler
        self.__copy = copy
        self.__run_level = run_level
        self.__write_to = better_mktemp()

    def __initiate_values(self):
        """
        Initiate all values.
        """
        self.__text_string = ''
        self.__field_instruction_string = ''
        self.__marker = 'mi<mk<inline-fld\n'
        self.__state = 'before_body'
        self.__string_obj = field_strings.FieldStrings(run_level=self.__run_level,
                bug_handler=self.__bug_handler,)
        self.__state_dict = {
        'before_body'       : self.__before_body_func,
        'in_body'           : self.__in_body_func,
        'field'             : self.__in_field_func,
        'field_instruction' : self.__field_instruction_func,
        }
        self.__in_body_dict = {
        'cw<fd<field_____'  : self.__found_field_func,
        }
        self.__field_dict = {
        'cw<fd<field-inst'  :   self.__found_field_instruction_func,
        'cw<fd<field_____'  : self.__found_field_func,
        'cw<pf<par-end___'  : self.__par_in_field_func,
        'cw<sc<section___'  : self.__sec_in_field_func,
        }
        self.__field_count = []  # keep track of the brackets
        self.__field_instruction = []  # field instruction strings
        self.__symbol = 0   # whether or not the field is really UTF-8
        # (these fields cannot be nested.)
        self.__field_instruction_string = ''  # string that collects field instruction
        self.__par_in_field = []  # paragraphs in field?
        self.__sec_in_field = []  # sections in field?
        self.__field_string = []  # list of field strings

    def __before_body_func(self, line):
        """
        Required:
            line --line ro parse
        Returns:
            nothing (changes an instant and writes a line)
        Logic:
            Check for the beginninf of the body. If found, changed the state.
            Always write out the line.
        """
        if self.__token_info == 'mi<mk<body-open_':
            self.__state = 'in_body'
        self.__write_obj.write(line)

    def __in_body_func(self, line):
        """
        Required:
            line --line to parse
        Returns:
            nothing. (Writes a line to the output file, or performs other actions.)
        Logic:
            Check of the beginning of a field. Always output the line.
        """
        action = self.__in_body_dict.get(self.__token_info)
        if action:
            action(line)
        self.__write_obj.write(line)

    def __found_field_func(self, line):
        """
        Requires:
            line --line to parse
        Returns:
            nothing
        Logic:
            Set the values for parsing the field. Four lists have to have
            items appended to them.
        """
        self.__state = 'field'
        self.__cb_count = 0
        ob_count = self.__ob_count
        self.__field_string.append('')
        self.__field_count.append(ob_count)
        self.__sec_in_field.append(0)
        self.__par_in_field.append(0)

    def __in_field_func(self, line):
        """
        Requires:
            line --line to parse
        Returns:
            nothing.
        Logic:
            Check for the end of the field; a paragraph break; a section break;
            the beginning of another field; or the beginning of the field
            instruction.
        """
        if self.__cb_count == self.__field_count[-1]:
            self.__field_string[-1] += line
            self.__end_field_func()
        else:
            action = self.__field_dict.get(self.__token_info)
            if action:
                action(line)
            else:
                self.__field_string[-1] += line

    def __par_in_field_func(self, line):
        """
        Requires:
            line --line to parse
        Returns:
            nothing
        Logic:
            Write the line to the output file and set the last item in the
            paragraph in field list to true.
        """
        self.__field_string[-1] += line
        self.__par_in_field[-1] = 1

    def __sec_in_field_func(self, line):
        """
        Requires:
            line --line to parse
        Returns:
            nothing
        Logic:
            Write the line to the output file and set the last item in the
            section in field list to true.
        """
        self.__field_string[-1] += line
        self.__sec_in_field[-1] = 1

    def __found_field_instruction_func(self, line):
        """
        Requires:
            line -- line to parse
        Returns:
            nothing
        Change the state to field instruction. Set the open bracket count of
        the beginning of this field so  you know when it ends. Set the closed
        bracket count to 0 so you don't prematureley exit this state.
        """
        self.__state = 'field_instruction'
        self.__field_instruction_count = self.__ob_count
        self.__cb_count = 0

    def __field_instruction_func(self, line):
        """
        Requires:
            line --line to parse
        Returns:
            nothing
        Logic:
            Collect all the lines until the end of the field is reached.
            Process these lines with the module rtr.field_strings.
            Check if the field instruction is 'Symbol' (really UTF-8).
        """
        if self.__cb_count == self.__field_instruction_count:
            # The closing bracket should be written, since the opening bracket
            # was written
            self.__field_string[-1] += line
            my_list = self.__string_obj.process_string(
                self.__field_instruction_string, 'field_instruction')
            instruction = my_list[2]
            self.__field_instruction.append(instruction)
            if my_list[0] == 'Symbol':
                self.__symbol = 1
            self.__state = 'field'
            self.__field_instruction_string = ''
        else:
            self.__field_instruction_string += line

    def __end_field_func(self):
        """
        Requires:
            nothing
        Returns:
            Nothing
        Logic:
            Pop the last values in the instructions list, the fields list, the
            paragraph list, and the section list.
            If the field is a symbol, do not write the tags <field></field>,
            since this field is really just UTF-8.
            If the field contains paragraph or section breaks, it is a
            field-block rather than just a field.
            Write the paragraph or section markers for later parsing of the
            file.
            If the filed list contains more strings, add the latest
            (processed) string to the last string in the list. Otherwise,
            write the string to the output file.
        """
        last_bracket = self.__field_count.pop()
        instruction = self.__field_instruction.pop()
        inner_field_string = self.__field_string.pop()
        sec_in_field = self.__sec_in_field.pop()
        par_in_field = self.__par_in_field.pop()
        # add a closing bracket, since the closing bracket is not included in
        # the field string
        if self.__symbol:
            inner_field_string = '%scb<nu<clos-brack<%s\n' % \
            (instruction, last_bracket)
        elif sec_in_field or par_in_field:
            inner_field_string = \
            'mi<mk<fldbkstart\n'\
            'mi<tg<open-att__<field-block<type>%s\n%s'\
            'mi<mk<fldbk-end_\n' \
            'mi<tg<close_____<field-block\n'\
            'mi<mk<fld-bk-end\n' \
            % (instruction, inner_field_string)
        # write a marker to show an inline field for later parsing
        else:
            inner_field_string = \
            '%s' \
            'mi<tg<open-att__<field<type>%s\n%s'\
            'mi<tg<close_____<field\n'\
            % (self.__marker, instruction, inner_field_string)
        if sec_in_field:
            inner_field_string = 'mi<mk<sec-fd-beg\n' + inner_field_string + \
            'mi<mk<sec-fd-end\n'
        if par_in_field:
            inner_field_string = 'mi<mk<par-in-fld\n' + inner_field_string
        if len(self.__field_string) == 0:
            self.__write_field_string(inner_field_string)
        else:
            self.__field_string[-1] += inner_field_string
        self.__symbol = 0

    def __write_field_string(self, the_string):
        self.__state = 'in_body'
        self.__write_obj.write(the_string)

    def fix_fields(self):
        """
        Requires:
            nothing
        Returns:
            nothing (changes the original file)
        Logic:
            Read one line in at a time. Determine what action to take based on
            the state. If the state is before the body, look for the
            beginning of the body.
            If the state is body, send the line to the body method.
        """
        self.__initiate_values()
        read_obj = open_for_read(self.__file)
        self.__write_obj = open_for_write(self.__write_to)
        line_to_read = 1
        while line_to_read:
            line_to_read = read_obj.readline()
            line = line_to_read
            self.__token_info = line[:16]
            if self.__token_info == 'ob<nu<open-brack':
                self.__ob_count = line[-5:-1]
            if self.__token_info == 'cb<nu<clos-brack':
                self.__cb_count = line[-5:-1]
            action = self.__state_dict.get(self.__state)
            if action is None:
                sys.stderr.write('no no matching state in module styles.py\n')
                sys.stderr.write(self.__state + '\n')
            action(line)
        read_obj.close()
        self.__write_obj.close()
        copy_obj = copy.Copy(bug_handler=self.__bug_handler)
        if self.__copy:
            copy_obj.copy_file(self.__write_to, "fields_large.data")
        copy_obj.rename(self.__write_to, self.__file)
        os.remove(self.__write_to)



#!/usr/bin/env python


__license__   = 'GPL v3'
__copyright__ = '2011, Kovid Goyal <kovid@kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

import os
import re
import traceback
from collections import defaultdict
from contextlib import suppress
from functools import partial

from qt.core import (
    QAbstractItemView,
    QApplication,
    QBrush,
    QColor,
    QCursor,
    QDialog,
    QDrag,
    QFont,
    QIcon,
    QLinearGradient,
    QMenu,
    QModelIndex,
    QPalette,
    QPen,
    QPoint,
    QPointF,
    QRect,
    QSize,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    Qt,
    QTimer,
    QToolTip,
    QTreeView,
    pyqtSignal,
)

from calibre import sanitize_file_name
from calibre.constants import config_dir
from calibre.ebooks.metadata import rating_to_stars
from calibre.gui2 import FunctionDispatcher, choose_files, config, empty_index, gprefs, pixmap_to_data, question_dialog, rating_font, safe_open_url
from calibre.gui2.complete2 import EditWithComplete
from calibre.gui2.dialogs.edit_category_notes import EditNoteDialog
from calibre.gui2.tag_browser.model import COUNT_ROLE, DRAG_IMAGE_ROLE, TAG_SEARCH_STATES, TagsModel, TagTreeItem, rename_only_in_vl_question
from calibre.gui2.widgets import EnLineEdit
from calibre.utils.icu import sort_key
from calibre.utils.serialize import json_loads


class TagDelegate(QStyledItemDelegate):  # {{{

    def __init__(self, tags_view):
        QStyledItemDelegate.__init__(self, tags_view)
        self.old_look = False
        self.rating_pat = re.compile(r'[%s]' % rating_to_stars(3, True))
        self.rating_font = QFont(rating_font())
        self.tags_view = tags_view
        self.links_icon = QIcon.ic('external-link.png')
        self.notes_icon = QIcon.ic('notes.png')
        self.blank_icon = QIcon()

    def draw_average_rating(self, item, style, painter, option, widget):
        rating = item.average_rating
        if rating is None:
            return
        r = style.subElementRect(QStyle.SubElement.SE_ItemViewItemDecoration, option, widget)
        icon = option.icon
        painter.save()
        nr = r.adjusted(0, 0, 0, 0)
        nr.setBottom(r.bottom()-int(r.height()*(rating/5.0)))
        painter.setClipRect(nr)
        bg = option.palette.window()
        if self.old_look:
            bg = option.palette.alternateBase() if option.features&QStyleOptionViewItem.ViewItemFeature.Alternate else option.palette.base()
        painter.fillRect(r, bg)
        style.proxy().drawPrimitive(QStyle.PrimitiveElement.PE_PanelItemViewItem, option, painter, widget)
        painter.setOpacity(0.3)
        icon.paint(painter, r, option.decorationAlignment, QIcon.Mode.Normal, QIcon.State.On)
        painter.restore()

    def draw_icon(self, style, painter, option, widget):
        r = style.subElementRect(QStyle.SubElement.SE_ItemViewItemDecoration, option, widget)
        icon = option.icon
        icon.paint(painter, r, option.decorationAlignment, QIcon.Mode.Normal, QIcon.State.On)

    def paint_text(self, painter, rect, flags, text, hover, option):
        painter.save()
        pen = painter.pen()
        if QApplication.instance().is_dark_theme:
            if hover:
                pen.setColor(QColor(Qt.GlobalColor.black))
            else:
                pen.setColor(option.palette.color(QPalette.ColorRole.WindowText))
        painter.setPen(pen)
        painter.drawText(rect, flags, text)
        painter.restore()

    def draw_text(self, style, painter, option, widget, index, item):
        tr = style.subElementRect(QStyle.SubElement.SE_ItemViewItemText, option, widget)
        text = index.data(Qt.ItemDataRole.DisplayRole)
        flags = Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignLeft | Qt.TextFlag.TextSingleLine
        lr = QRect(tr)
        lr.setRight(lr.right() * 2)
        text_rec = painter.boundingRect(lr, flags, text)
        hover = option.state & QStyle.StateFlag.State_MouseOver
        is_search = (True if item.type == TagTreeItem.TAG and
                            item.tag.category == 'search' else False)

        def render_count():
            if not is_search and (hover or gprefs['tag_browser_show_counts']):
                count = str(index.data(COUNT_ROLE))
                width = painter.fontMetrics().boundingRect(count).width()
                r = QRect(tr)
                r.setRight(r.right() - 1), r.setLeft(r.right() - width - 4)
                self.paint_text(painter, r, Qt.AlignmentFlag.AlignCenter | Qt.TextFlag.TextSingleLine, count, hover, option)
                tr.setRight(r.left() - 1)
            else:
                tr.setRight(tr.right() - 1)

        if item.type == TagTreeItem.TAG:
            category = item.tag.category
            name = item.tag.original_name
            tv = self.tags_view
            m = tv._model
            positions = {'links': (-1, -1), 'notes': (-1, -1)}

            # The icons fits in a rectangle height/2 + 4 x height/2 + 4. This
            # ensures they are a 'pleasant' size compared to the text.
            icon_width = int(tr.height()/2) + 4

            def render_link_icon():
                icon = self.links_icon if m.item_has_link(category, name) else self.blank_icon
                r = QRect(tr)
                r.setRight(r.right() - 1)
                r.setLeft(r.right() - icon_width)
                positions['links'] = (r.left(), r.left()+r.width())
                icon.paint(painter, r, option.decorationAlignment, QIcon.Mode.Normal, QIcon.State.On)
                tr.setRight(r.left() - 1)
            def render_note_icon():
                icon = self.notes_icon if m.item_has_note(category, name) else self.blank_icon
                r = QRect(tr)
                r.setRight(r.right() - 1)
                r.setLeft(r.right() - icon_width)
                positions['notes'] = (r.left(), r.left()+r.width())
                icon.paint(painter, r, option.decorationAlignment, QIcon.Mode.Normal, QIcon.State.On)
                tr.setRight(r.left() - 1)

            if gprefs['icons_on_right_in_tag_browser']:
                # Icons go far right, in columns after the counts
                show_note_icon = gprefs['show_notes_in_tag_browser'] and m.category_has_notes(category)
                show_link_icon = gprefs['show_links_in_tag_browser'] and m.category_has_links(category)
                if show_link_icon:
                    render_link_icon()
                if show_note_icon:
                    render_note_icon()
                render_count()
            else:
                # Icons go after the text to the left of the counts, not in columns
                show_note_icon = gprefs['show_notes_in_tag_browser'] and m.item_has_note(category, name)
                show_link_icon = gprefs['show_links_in_tag_browser'] and m.item_has_link(category, name)

                render_count()
                # The link icon has a margin of 1 px on each side. Account for
                # this when computing the width of the icons. If you change the
                # order of the icons then you must change this calculation
                w = (int(show_link_icon) * (icon_width + 2)) + (int(show_note_icon) * icon_width)
                # Leave a 5 px margin between the text and the icon.
                tr.setWidth(min(tr.width(), text_rec.width() + 5 + w))
                if show_link_icon:
                    render_link_icon()
                if show_note_icon:
                    render_note_icon()
            tv.category_button_positions[category][name] = positions
        else:
            render_count()

        is_rating = item.type == TagTreeItem.TAG and not self.rating_pat.sub('', text)
        if is_rating:
            painter.setFont(self.rating_font)
        if text_rec.width() > tr.width():
            g = QLinearGradient(QPointF(tr.topLeft()), QPointF(tr.topRight()))
            c = option.palette.color(QPalette.ColorRole.WindowText)
            g.setColorAt(0, c), g.setColorAt(0.8, c)
            c = QColor(c)
            c.setAlpha(0)
            g.setColorAt(1, c)
            pen = QPen()
            pen.setBrush(QBrush(g))
            painter.setPen(pen)
        self.paint_text(painter, tr, flags, text, hover, option)

    def paint(self, painter, option, index):
        QStyledItemDelegate.paint(self, painter, option, empty_index)
        widget = self.parent()
        style = QApplication.style() if widget is None else widget.style()
        self.initStyleOption(option, index)
        item = index.data(Qt.ItemDataRole.UserRole)
        self.draw_icon(style, painter, option, widget)
        painter.save()
        self.draw_text(style, painter, option, widget, index, item)
        painter.restore()
        if item.boxed:
            r = style.subElementRect(QStyle.SubElement.SE_ItemViewItemFocusRect, option,
                    widget)
            painter.drawLine(r.bottomLeft(), r.bottomRight())
        if item.type == TagTreeItem.TAG and item.tag.state == 0 and config['show_avg_rating']:
            self.draw_average_rating(item, style, painter, option, widget)

    def createEditor(self, parent, option, index):
        item = self.tags_view.model().get_node(index)
        if not item.ignore_vl:
            if item.use_vl is None:
                if self.tags_view.model().get_in_vl():
                    item.use_vl = rename_only_in_vl_question(self.tags_view)
                else:
                    item.use_vl = False
            elif not item.use_vl and self.tags_view.model().get_in_vl():
                item.use_vl = not question_dialog(self.tags_view,
                                    _('Rename in Virtual library'), '<p>' +
                                    _('A Virtual library is active but you are renaming '
                                      'the item in all books in your library. Is '
                                      'this really what you want to do?') + '</p>',
                                    yes_text=_('Yes, apply in entire library'),
                                    no_text=_('No, apply only in Virtual library'),
                                    skip_dialog_name='tag_item_rename_in_entire_library')
        key, completion_data = '', None
        if item.type == TagTreeItem.CATEGORY:
            key = item.category_key
        elif item.type == TagTreeItem.TAG:
            key = getattr(item.tag, 'category', '')
        if key:
            from calibre.gui2.ui import get_gui
            with suppress(Exception):
                completion_data = get_gui().current_db.new_api.all_field_names(key)
        if completion_data:
            editor = EditWithComplete(parent)
            editor.set_separator(None)
            editor.update_items_cache(completion_data)
        else:
            editor = EnLineEdit(parent)
        return editor

    # }}}


class TagsView(QTreeView):  # {{{

    refresh_required        = pyqtSignal()
    tags_marked             = pyqtSignal(object)
    edit_user_category      = pyqtSignal(object)
    delete_user_category    = pyqtSignal(object)
    del_item_from_user_cat  = pyqtSignal(object, object, object)
    add_item_to_user_cat    = pyqtSignal(object, object, object)
    add_subcategory         = pyqtSignal(object)
    tags_list_edit          = pyqtSignal(object, object, object)
    saved_search_edit       = pyqtSignal(object)
    rebuild_saved_searches  = pyqtSignal()
    author_sort_edit        = pyqtSignal(object, object, object, object, object)
    tag_item_renamed        = pyqtSignal()
    search_item_renamed     = pyqtSignal()
    drag_drop_finished      = pyqtSignal(object)
    restriction_error       = pyqtSignal(object)
    tag_item_delete         = pyqtSignal(object, object, object, object, object)
    tag_identifier_delete   = pyqtSignal(object, object)
    apply_tag_to_selected   = pyqtSignal(object, object, object)
    edit_enum_values        = pyqtSignal(object, object, object)

    def __init__(self, parent=None):
        QTreeView.__init__(self, parent=None)
        self.possible_drag_start = None
        self.setProperty('frame_for_focus', True)
        self.setMouseTracking(True)
        self.alter_tb = None
        self.disable_recounting = False
        self.setUniformRowHeights(True)
        self.setIconSize(QSize(20, 20))
        self.setTabKeyNavigation(True)
        self.setAnimated(True)
        self.setHeaderHidden(True)
        self.setItemDelegate(TagDelegate(tags_view=self))
        self.made_connections = False
        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.setDropIndicatorShown(True)
        self.setAutoExpandDelay(500)
        self.pane_is_visible = False
        self.search_icon = QIcon.ic('search.png')
        self.search_copy_icon = QIcon.ic("search_copy_saved.png")
        self.user_category_icon = QIcon.ic('tb_folder.png')
        self.edit_metadata_icon = QIcon.ic('edit_input.png')
        self.delete_icon = QIcon.ic('list_remove.png')
        self.rename_icon = QIcon.ic('edit-undo.png')
        self.plus_icon = QIcon.ic('plus.png')
        self.minus_icon = QIcon.ic('minus.png')

        # Dict for recording the positions of the fake buttons for category tag
        # lines. It is recorded per category because we can't guarantee the
        # order that items are painted. The numbers get updated whenever an item
        # is painted, which deals with resizing.
        self.category_button_positions = defaultdict(dict)

        self._model = TagsModel(self)
        self._model.search_item_renamed.connect(self.search_item_renamed)
        self._model.refresh_required.connect(self.refresh_required,
                type=Qt.ConnectionType.QueuedConnection)
        self._model.tag_item_renamed.connect(self.tag_item_renamed)
        self._model.restriction_error.connect(self.restriction_error)
        self._model.user_categories_edited.connect(self.user_categories_edited,
                type=Qt.ConnectionType.QueuedConnection)
        self._model.drag_drop_finished.connect(self.drag_drop_finished)
        self._model.convert_requested.connect(self.convert_requested)
        self.set_look_and_feel(first=True)
        QApplication.instance().palette_changed.connect(self.set_style_sheet, type=Qt.ConnectionType.QueuedConnection)
        self.marked_change_listener = FunctionDispatcher(self.recount_on_mark_change)

    def convert_requested(self, book_ids, to_fmt):
        from calibre.gui2.ui import get_gui
        get_gui().iactions['Convert Books'].convert_ebooks_to_format(book_ids, to_fmt)

    def set_style_sheet(self):
        stylish_tb = '''
                QTreeView {
                    background-color: palette(window);
                    color: palette(window-text);
                    border: none;
                }
        '''
        self.setStyleSheet('''
                QTreeView::item {
                    border: 1px solid transparent;
                    padding-top:PADex;
                    padding-bottom:PADex;
                }

        '''.replace('PAD', str(gprefs['tag_browser_item_padding'])) + (
            '' if gprefs['tag_browser_old_look'] else stylish_tb) + QApplication.instance().palette_manager.tree_view_hover_style()
        )
        self.setProperty('hovered_item_is_highlighted', True)

    def set_look_and_feel(self, first=False):
        self.set_style_sheet()
        self.setAlternatingRowColors(gprefs['tag_browser_old_look'])
        self.itemDelegate().old_look = gprefs['tag_browser_old_look']

        if gprefs['tag_browser_allow_keyboard_focus']:
            self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        else:
            self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        # Ensure the TB doesn't keep the focus it might already have. When this
        # method is first called during GUI initialization not everything is
        # set up, in which case don't try to change the focus.
        # Note: this process has the side effect of moving the focus to the
        # library view whenever a look & feel preference is changed.
        if not first:
            try:
                from calibre.gui2.ui import get_gui
                get_gui().shift_esc()
            except:
                traceback.print_exc()

    @property
    def hidden_categories(self):
        return self._model.hidden_categories

    @property
    def db(self):
        return self._model.db

    @property
    def collapse_model(self):
        return self._model.collapse_model

    def set_pane_is_visible(self, to_what):
        pv = self.pane_is_visible
        self.pane_is_visible = to_what
        if to_what and not pv:
            self.recount()

    def get_state(self):
        state_map = {}
        expanded_categories = []
        hide_empty_categories = self.model().prefs['tag_browser_hide_empty_categories']
        crmap = self._model.category_row_map()
        for category in self._model.category_nodes:
            if (category.category_key in self.hidden_categories or (
                hide_empty_categories and len(category.child_tags()) == 0)):
                continue
            row = crmap.get(category.category_key)
            if row is not None:
                index = self._model.index(row, 0, QModelIndex())
                if self.isExpanded(index):
                    expanded_categories.append(category.category_key)
            states = [c.tag.state for c in category.child_tags()]
            names = [(c.tag.name, c.tag.category) for c in category.child_tags()]
            state_map[category.category_key] = dict(zip(names, states))
        return expanded_categories, state_map

    def reread_collapse_parameters(self):
        self._model.reread_collapse_model(self.get_state()[1])

    def set_database(self, db, alter_tb):
        self._model.set_database(db)
        self.alter_tb = alter_tb
        self.pane_is_visible = True  # because TagsModel.set_database did a recount
        self.setModel(self._model)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        pop = self.db.CATEGORY_SORTS.index(config['sort_tags_by'])
        self.alter_tb.sort_menu.actions()[pop].setChecked(True)
        try:
            match_pop = self.db.MATCH_TYPE.index(config['match_tags_type'])
        except ValueError:
            match_pop = 0
        self.alter_tb.match_menu.actions()[match_pop].setChecked(True)
        if not self.made_connections:
            self.clicked.connect(self.toggle_on_mouse_click)
            self.customContextMenuRequested.connect(self.show_context_menu)
            self.refresh_required.connect(self.recount, type=Qt.ConnectionType.QueuedConnection)
            self.alter_tb.sort_menu.triggered.connect(self.sort_changed)
            self.alter_tb.match_menu.triggered.connect(self.match_changed)
            self.made_connections = True
        self.refresh_signal_processed = True
        db.add_listener(self.database_changed)
        self.expanded.connect(self.item_expanded)
        self.collapsed.connect(self.collapse_node_and_children)
        db.data.add_marked_listener(self.marked_change_listener)

    def keyPressEvent(self, event):

        def on_last_visible_item(dex, check_children):
            model = self._model
            if model.get_node(dex) == model.root_item:
                # Got to root. There can't be any more children to show
                return True
            if check_children and self.isExpanded(dex):
                # We are on a node with expanded children so there is a node to go to.
                # We don't check children if we are moving up the parent hierarchy
                return False
            parent = model.parent(dex)
            if dex.row() < model.rowCount(parent) - 1:
                # Node has more nodes after it
                return False
            # Last node. Check the parent for further to see if there are more nodes
            return on_last_visible_item(parent, False)

        # I don't see how current_index can ever be not valid, but ...
        if self.currentIndex().isValid():
            key = event.key()
            if gprefs['tag_browser_allow_keyboard_focus']:
                if key == Qt.Key.Key_Return and self.state() != QAbstractItemView.State.EditingState:
                    self.toggle_current_index()
                    return
                # Check if we are moving the focus and we are at the beginning or the
                # end of the list. The goal is to prevent moving focus away from the
                # tag browser.
                if key == Qt.Key.Key_Tab:
                    if not on_last_visible_item(self.currentIndex(), True):
                        QTreeView.keyPressEvent(self, event)
                    return
                if key == Qt.Key.Key_Backtab:
                    if self.model().get_node(self.currentIndex()) != self._model.root_item.children[0]:
                        QTreeView.keyPressEvent(self, event)
                    return
            # If this is an edit request, mark the node to request whether to use VLs
            # As far as I can tell, F2 is used across all platforms
            if key == Qt.Key.Key_F2:
                node = self.model().get_node(self.currentIndex())
                if node.type == TagTreeItem.TAG:
                    # Saved search nodes don't use the VL test/dialog
                    node.use_vl = None
                    node.ignore_vl = node.tag.category == 'search'
                else:
                    # Don't open the editor for non-editable items
                    if not node.category_key.startswith('@') or node.is_gst:
                        return
                    # Category nodes don't use the VL test/dialog
                    node.use_vl = False
                    node.ignore_vl = True
        QTreeView.keyPressEvent(self, event)

    def database_changed(self, event, ids):
        if self.refresh_signal_processed:
            self.refresh_signal_processed = False
            self.refresh_required.emit()

    def user_categories_edited(self, user_cats, nkey):
        state_map = self.get_state()[1]
        self.db.new_api.set_pref('user_categories', user_cats)
        self._model.rebuild_node_tree(state_map=state_map)
        p = self._model.find_category_node('@'+nkey)
        self.show_item_at_path(p)

    @property
    def match_all(self):
        return (self.alter_tb and self.alter_tb.match_menu.actions()[1].isChecked())

    def sort_changed(self, action):
        for i, ac in enumerate(self.alter_tb.sort_menu.actions()):
            if ac is action:
                config.set('sort_tags_by', self.db.CATEGORY_SORTS[i])
                self.recount()
                break

    def match_changed(self, action):
        try:
            for i, ac in enumerate(self.alter_tb.match_menu.actions()):
                if ac is action:
                    config.set('match_tags_type', self.db.MATCH_TYPE[i])
        except:
            pass

    def mousePressEvent(self, event):
        if event.buttons() & Qt.MouseButton.LeftButton:
            # Record the press point for processing during the clicked signal
            self.mouse_clicked_point = event.pos()
            # Only remember a possible drag start if the item is drag enabled
            dex = self.indexAt(event.pos())
            if self._model.flags(dex) & Qt.ItemFlag.ItemIsDragEnabled:
                self.possible_drag_start = event.pos()
            else:
                self.possible_drag_start = None
        return QTreeView.mousePressEvent(self, event)

    def mouseMoveEvent(self, event):
        dex = self.indexAt(event.pos())
        if dex.isValid():
            self.setCursor(Qt.CursorShape.PointingHandCursor)
        else:
            self.unsetCursor()
        if not event.buttons() & Qt.MouseButton.LeftButton:
            return
        if not dex.isValid():
            QTreeView.mouseMoveEvent(self, event)
            return
        # don't start drag/drop until the mouse has moved a bit.
        if (self.possible_drag_start is None or
            (event.pos() - self.possible_drag_start).manhattanLength() <
                                    QApplication.startDragDistance()):
            QTreeView.mouseMoveEvent(self, event)
            return

        if not self._model.flags(dex) & Qt.ItemFlag.ItemIsDragEnabled:
            QTreeView.mouseMoveEvent(self, event)
            return
        md = self._model.mimeData([dex])
        pixmap = dex.data(DRAG_IMAGE_ROLE).pixmap(self.iconSize())
        drag = QDrag(self)
        drag.setPixmap(pixmap)
        drag.setMimeData(md)
        if (self._model.is_in_user_category(dex) or
                    self._model.is_index_on_a_hierarchical_category(dex)):
            '''
            Things break if we specify MoveAction as the default, which is
            what we want for drag on hierarchical categories. Dragging user
            categories stops working. Don't know why. To avoid the problem
            we fix the action in dragMoveEvent.
            '''
            drag.exec(Qt.DropAction.CopyAction|Qt.DropAction.MoveAction, Qt.DropAction.CopyAction)
        else:
            drag.exec(Qt.DropAction.CopyAction)

    def mouseDoubleClickEvent(self, event):
        # swallow these to avoid toggling and editing at the same time
        pass

    @property
    def search_string(self):
        tokens = self._model.tokens()
        joiner = ' and ' if self.match_all else ' or '
        return joiner.join(tokens)

    def click_in_button_range(self, val, category, item_name, kind):
        range_tuple = self.category_button_positions[category].get(item_name, {}).get(kind)
        return range_tuple and range_tuple[0] <= val <= range_tuple[1]

    def toggle_current_index(self):
        ci = self.currentIndex()
        if ci.isValid():
            self.toggle(ci)

    def toggle_on_mouse_click(self, index):
        # Check if one of the link or note icons was clicked. If so, deal with
        # it here and don't do the real toggle
        t = self._model.data(index, Qt.UserRole)
        if t.type == TagTreeItem.TAG:
            db = self._model.db.new_api
            category = t.tag.category
            orig_name = t.tag.original_name
            x = self.mouse_clicked_point.x()
            if self.click_in_button_range(x, category, orig_name, 'notes'):
                from calibre.gui2.dialogs.show_category_note import ShowNoteDialog
                item_id = db.get_item_id(category, orig_name)
                if db.notes_for(category, item_id):
                    ShowNoteDialog(category, item_id, db, parent=self).show()
                    return
            if self.click_in_button_range(x, category, orig_name, 'links'):
                link = db.get_link_map(category).get(orig_name)
                if link:
                    safe_open_url(link)
                    return
        self._toggle(index, None)

    def toggle(self, index):
        self._toggle(index, None)

    def _toggle(self, index, set_to):
        '''
        set_to: if None, advance the state. Otherwise must be one of the values
        in TAG_SEARCH_STATES
        '''
        exclusive = QApplication.keyboardModifiers() not in (Qt.KeyboardModifier.ControlModifier, Qt.KeyboardModifier.ShiftModifier)
        if self._model.toggle(index, exclusive, set_to=set_to):
            # Reset the focus back to TB if it has it before the toggle
            # Must ask this question before starting the search because
            # it changes the focus
            has_focus = self.hasFocus()
            self.tags_marked.emit(self.search_string)
            if has_focus and gprefs['tag_browser_allow_keyboard_focus']:
                # Reset the focus to the TB. Use the singleshot in case
                # some of searching is done using queued signals.
                QTimer.singleShot(0, lambda: self.setFocus())

    def conditional_clear(self, search_string):
        if search_string != self.search_string:
            self.clear()

    def context_menu_handler(self, action=None, category=None,
                             key=None, index=None, search_state=None,
                             is_first_letter=False, ignore_vl=False,
                             extra=None):
        if not action:
            return
        from calibre.gui2.ui import get_gui
        try:
            if action == 'edit_note':
                if EditNoteDialog(category, extra, self.db).exec() == QDialog.DialogCode.Accepted:
                    get_gui().do_field_item_value_changed()
                return
            if action == 'dont_collapse_category':
                if key not in extra:
                    extra.append(key)
                self.db.prefs.set('tag_browser_dont_collapse', extra)
                self.recount()
                return
            if action == 'collapse_category':
                if key in extra:
                    extra.remove(key)
                self.db.prefs.set('tag_browser_dont_collapse', extra)
                self.recount()
                return
            if action == 'set_icon':
                try:
                    path = choose_files(self, 'choose_category_icon',
                                _('Change icon for: %s')%key, filters=[
                                ('Images', ['png', 'gif', 'jpg', 'jpeg'])],
                            all_files=False, select_only_single_file=True)
                    if path:
                        path = path[0]
                        p = QIcon(path).pixmap(QSize(128, 128))
                        d = os.path.join(config_dir, 'tb_icons')
                        if not os.path.exists(d):
                            os.makedirs(d)
                        with open(os.path.join(d, 'icon_' + sanitize_file_name(key)+'.png'), 'wb') as f:
                            f.write(pixmap_to_data(p, format='PNG'))
                            path = os.path.basename(f.name)
                        self._model.set_custom_category_icon(key, str(path))
                        self.recount()
                except:
                    traceback.print_exc()
                return
            if action == 'clear_icon':
                self._model.set_custom_category_icon(key, None)
                self.recount()
                return

            if action == 'edit_item_no_vl':
                item = self.model().get_node(index)
                item.use_vl = False
                item.ignore_vl = ignore_vl
                self.edit(index)
                return
            if action == 'edit_item_in_vl':
                item = self.model().get_node(index)
                item.use_vl = True
                item.ignore_vl = ignore_vl
                self.edit(index)
                return
            if action == 'delete_item_in_vl':
                tag = index.tag
                id_ = tag.id if tag.is_editable else None
                children = index.child_tags()
                self.tag_item_delete.emit(key, id_, tag.original_name,
                                          self.model().get_book_ids_to_use(),
                                          children)
                return
            if action == 'delete_item_no_vl':
                tag = index.tag
                id_ = tag.id if tag.is_editable else None
                children = index.child_tags()
                self.tag_item_delete.emit(key, id_, tag.original_name,
                                          None, children)
                return
            if action == 'delete_identifier':
                self.tag_identifier_delete.emit(index.tag.name, False)
                return
            if action == 'delete_identifier_in_vl':
                self.tag_identifier_delete.emit(index.tag.name, True)
                return
            if action == 'open_editor':
                self.tags_list_edit.emit(category, key, is_first_letter)
                return
            if action == 'manage_categories':
                self.edit_user_category.emit(category)
                return
            if action == 'search':
                self._toggle(index, set_to=search_state)
                return
            if action == "raw_search":
                get_gui().get_saved_search_text(search_name='search:' + key)
                return
            if action == 'add_to_category':
                tag = index.tag
                if len(index.children) > 0:
                    for c in index.all_children():
                        self.add_item_to_user_cat.emit(category, c.tag.original_name,
                                               c.tag.category)
                self.add_item_to_user_cat.emit(category, tag.original_name,
                                               tag.category)
                return
            if action == 'add_subcategory':
                self.add_subcategory.emit(key)
                return
            if action == 'search_category':
                self._toggle(index, set_to=search_state)
                return
            if action == 'delete_user_category':
                self.delete_user_category.emit(key)
                return
            if action == 'delete_search':
                if not question_dialog(
                    self,
                    title=_('Delete Saved search'),
                    msg='<p>'+ _('Delete the saved search: {}?').format(key),
                    skip_dialog_name='tb_delete_saved_search',
                    skip_dialog_msg=_('Show this confirmation again')
                ):
                    return
                self.model().db.saved_search_delete(key)
                self.rebuild_saved_searches.emit()
                return
            if action == 'delete_item_from_user_category':
                tag = index.tag
                if len(index.children) > 0:
                    for c in index.children:
                        self.del_item_from_user_cat.emit(key, c.tag.original_name,
                                               c.tag.category)
                self.del_item_from_user_cat.emit(key, tag.original_name, tag.category)
                return
            if action == 'manage_searches':
                self.saved_search_edit.emit(category)
                return
            if action == 'edit_authors':
                self.author_sort_edit.emit(self, index, False, False, is_first_letter)
                return
            if action == 'edit_author_sort':
                self.author_sort_edit.emit(self, index, True, False, is_first_letter)
                return
            if action == 'edit_author_link':
                self.author_sort_edit.emit(self, index, False, True, False)
                return
            if action == 'remove_format':
                gui = get_gui()
                gui.iactions['Remove Books'].remove_format_from_selected_books(key)
                return
            if action == 'edit_open_with_apps':
                from calibre.gui2.open_with import edit_programs
                edit_programs(key, self)
                return
            if action == 'add_open_with_apps':
                from calibre.gui2.open_with import choose_program
                choose_program(key, self)
                return

            reset_filter_categories = True
            if action == 'hide':
                self.hidden_categories.add(category)
            elif action == 'show':
                self.hidden_categories.discard(category)
            elif action == 'categorization':
                changed = self.collapse_model != category
                self._model.collapse_model = category
                if changed:
                    reset_filter_categories = False
                    gprefs['tags_browser_partition_method'] = category
            elif action == 'defaults':
                self.hidden_categories.clear()
            elif action == 'add_tag':
                item = self.model().get_node(index)
                if item is not None:
                    self.apply_to_selected_books(item)
                return
            elif action == 'remove_tag':
                item = self.model().get_node(index)
                if item is not None:
                    self.apply_to_selected_books(item, True)
                return
            elif action == 'edit_enum':
                self.edit_enum_values.emit(self, self.db, key)
                return
            self.db.new_api.set_pref('tag_browser_hidden_categories', list(self.hidden_categories))
            if reset_filter_categories:
                self._model.set_categories_filter(None)
            self._model.rebuild_node_tree()
        except Exception:
            traceback.print_exc()
            return

    def apply_to_selected_books(self, item, remove=False):
        if item.type != item.TAG:
            return
        tag = item.tag
        if not tag.category or not tag.original_name:
            return
        self.apply_tag_to_selected.emit(tag.category, tag.original_name, remove)

    def show_context_menu(self, point):
        def display_name(tag):
            ans = tag.name
            if tag.category == 'search':
                n = tag.name
                if len(n) > 45:
                    n = n[:45] + '...'
                ans = n
            elif tag.is_hierarchical and not tag.is_editable:
                ans = tag.original_name
            if ans:
                ans = ans.replace('&', '&&')
            return ans

        index = self.indexAt(point)
        self.context_menu = QMenu(self)
        added_show_hidden_categories = False
        key = None

        def add_show_hidden_categories():
            nonlocal added_show_hidden_categories
            if self.hidden_categories and not added_show_hidden_categories:
                added_show_hidden_categories = True
                m = self.context_menu.addMenu(_('Show category'))
                m.setIcon(QIcon.ic('plus.png'))
                # The search category can disappear from field_metadata. Perhaps
                # other dynamic categories can as well. The implication is that
                # dynamic categories are being removed, but how that would
                # happen is a mystery. I suspect a plugin is operating on the
                # "real" field_metadata instead of a copy, thereby changing the
                # dict used by the rest of calibre.
                #
                # As it can happen, to avoid key errors check that a category
                # exists before offering to unhide it.
                for col in sorted((c for c in self.hidden_categories if c in self.db.field_metadata),
                        key=lambda x: sort_key(self.db.field_metadata[x]['name'])):
                    ac = m.addAction(self.db.field_metadata[col]['name'],
                        partial(self.context_menu_handler, action='show', category=col))
                    ic = self.model().category_custom_icons.get(col)
                    if ic:
                        ac.setIcon(QIcon.ic(ic))
                m.addSeparator()
                m.addAction(_('All categories'),
                        partial(self.context_menu_handler, action='defaults')).setIcon(QIcon.ic('plusplus.png'))

        search_submenu = None
        if index.isValid():
            item = index.data(Qt.ItemDataRole.UserRole)
            tag = None
            tag_item = item

            if item.type == TagTreeItem.TAG:
                tag = item.tag
                while item.type != TagTreeItem.CATEGORY:
                    item = item.parent

            if item.type == TagTreeItem.CATEGORY:
                if not item.category_key.startswith('@'):
                    while item.parent != self._model.root_item:
                        item = item.parent
                category = str(item.name or '')
                key = item.category_key
                # Verify that we are working with a field that we know something about
                if key not in self.db.field_metadata:
                    return True
                fm = self.db.field_metadata[key]

                # Did the user click on a leaf node?
                if tag:
                    # If the user right-clicked on an editable item, then offer
                    # the possibility of renaming that item.
                    if (fm['datatype'] != 'composite' and
                            (tag.is_editable or tag.is_hierarchical) and
                            key != 'search'):
                        # Add the 'rename' items to both interior and leaf nodes
                        if fm['datatype'] != 'enumeration':
                            if self.model().get_in_vl():
                                self.context_menu.addAction(self.rename_icon,
                                        _('Rename %s in Virtual library')%display_name(tag),
                                        partial(self.context_menu_handler, action='edit_item_in_vl',
                                                index=index, category=key))
                            self.context_menu.addAction(self.rename_icon,
                                        _('Rename %s')%display_name(tag),
                                        partial(self.context_menu_handler, action='edit_item_no_vl',
                                                index=index, category=key))
                        if key in ('tags', 'series', 'publisher') or \
                                self._model.db.field_metadata.is_custom_field(key):
                            if self.model().get_in_vl():
                                self.context_menu.addAction(self.delete_icon,
                                                    _('Delete %s in Virtual library')%display_name(tag),
                                partial(self.context_menu_handler, action='delete_item_in_vl',
                                    key=key, index=tag_item))

                            self.context_menu.addAction(self.delete_icon,
                                                    _('Delete %s')%display_name(tag),
                                partial(self.context_menu_handler, action='delete_item_no_vl',
                                    key=key, index=tag_item))
                    if tag.is_editable:
                        if key == 'authors':
                            self.context_menu.addAction(_('Edit sort for %s')%display_name(tag),
                                    partial(self.context_menu_handler,
                                            action='edit_author_sort', index=tag.id)).setIcon(QIcon.ic('auto_author_sort.png'))
                            self.context_menu.addAction(_('Edit link for %s')%display_name(tag),
                                    partial(self.context_menu_handler,
                                            action='edit_author_link', index=tag.id)).setIcon(QIcon.ic('insert-link.png'))
                        elif self.db.new_api.has_link_map(key):
                            self.context_menu.addAction(_('Edit link for %s')%display_name(tag),
                                    partial(self.context_menu_handler, action='open_editor',
                                            category=tag.original_name if tag else None,
                                            key=key))

                        if self.db.new_api.field_supports_notes(key):
                            item_id = self.db.new_api.get_item_id(tag.category, tag.original_name)
                            has_note = self._model.item_has_note(key, tag.original_name)
                            self.context_menu.addAction(self.edit_metadata_icon,
                                (_('Edit note for %s') if has_note else _('Create note for %s'))%display_name(tag),
                                partial(self.context_menu_handler, action='edit_note',
                                        index=index, extra=item_id, category=tag.category))

                        # is_editable is also overloaded to mean 'can be added
                        # to a User category'
                        m = QMenu(_('Add %s to User category')%display_name(tag), self.context_menu)
                        m.setIcon(self.user_category_icon)
                        added = [False]

                        def add_node_tree(tree_dict, m, path):
                            p = path[:]
                            for k in sorted(tree_dict.keys(), key=sort_key):
                                p.append(k)
                                n = k[1:] if k.startswith('@') else k
                                m.addAction(self.user_category_icon, n,
                                    partial(self.context_menu_handler,
                                            'add_to_category',
                                            category='.'.join(p), index=tag_item))
                                added[0] = True
                                if len(tree_dict[k]):
                                    tm = m.addMenu(self.user_category_icon,
                                                   _('Children of %s')%n)
                                    add_node_tree(tree_dict[k], tm, p)
                                p.pop()
                        add_node_tree(self.model().user_category_node_tree, m, [])
                        if added[0]:
                            self.context_menu.addMenu(m)

                        # is_editable also means the tag can be applied/removed
                        # from selected books
                        if fm['datatype'] != 'rating':
                            m = self.context_menu.addMenu(self.edit_metadata_icon,
                                            _('Add/remove %s to selected books')%display_name(tag))
                            m.addAction(self.plus_icon,
                                _('Add %s to selected books') % display_name(tag),
                                partial(self.context_menu_handler, action='add_tag', index=index))
                            m.addAction(self.minus_icon,
                                _('Remove %s from selected books') % display_name(tag),
                                partial(self.context_menu_handler, action='remove_tag', index=index))
                    elif key == 'search' and tag.is_searchable:
                        self.context_menu.addAction(self.rename_icon,
                                                    _('Rename %s')%display_name(tag),
                            partial(self.context_menu_handler, action='edit_item_no_vl',
                                    index=index, ignore_vl=True))
                        self.context_menu.addAction(self.delete_icon,
                                _('Delete Saved search %s')%display_name(tag),
                                partial(self.context_menu_handler,
                                        action='delete_search', key=tag.original_name))
                    elif key == 'identifiers':
                        if self.model().get_in_vl():
                            self.context_menu.addAction(self.delete_icon,
                                    _('Delete %s in Virtual Library')%display_name(tag),
                                    partial(self.context_menu_handler,
                                            action='delete_identifier_in_vl',
                                            key=key, index=tag_item))
                        else:
                            self.context_menu.addAction(self.delete_icon,
                                    _('Delete %s')%display_name(tag),
                                    partial(self.context_menu_handler,
                                            action='delete_identifier',
                                            key=key, index=tag_item))

                    if key.startswith('@') and not item.is_gst:
                        self.context_menu.addAction(self.user_category_icon,
                            _('Remove {item} from category: {cat}').format(item=display_name(tag), cat=item.py_name),
                            partial(self.context_menu_handler, action='delete_item_from_user_category', key=key, index=tag_item))
                    if tag.is_searchable:
                        # Add the search for value items. All leaf nodes are searchable
                        self.context_menu.addSeparator()
                        search_submenu = self.context_menu.addMenu(_('Search for'))
                        search_submenu.setIcon(QIcon.ic('search.png'))
                        search_submenu.addAction(self.search_icon,
                                '%s'%display_name(tag),
                                partial(self.context_menu_handler, action='search',
                                        search_state=TAG_SEARCH_STATES['mark_plus'],
                                        index=index))
                        add_child_search = (tag.is_hierarchical == '5state' and
                                            len(tag_item.children))
                        if add_child_search:
                            search_submenu.addAction(self.search_icon,
                                    _('%s and its children')%display_name(tag),
                                    partial(self.context_menu_handler, action='search',
                                            search_state=TAG_SEARCH_STATES['mark_plusplus'],
                                            index=index))
                        search_submenu.addAction(self.search_icon,
                                _('Everything but %s')%display_name(tag),
                                partial(self.context_menu_handler, action='search',
                                        search_state=TAG_SEARCH_STATES['mark_minus'],
                                        index=index))
                        if add_child_search:
                            search_submenu.addAction(self.search_icon,
                                    _('Everything but %s and its children')%display_name(tag),
                                    partial(self.context_menu_handler, action='search',
                                            search_state=TAG_SEARCH_STATES['mark_minusminus'],
                                            index=index))
                        if key == 'search':
                            search_submenu.addAction(self.search_copy_icon,
                                     _('The saved search expression'),
                                     partial(self.context_menu_handler, action='raw_search',
                                             key=tag.original_name))
                    self.context_menu.addSeparator()
                elif key.startswith('@') and not item.is_gst:
                    if item.can_be_edited:
                        self.context_menu.addAction(self.rename_icon,
                            _('Rename %s')%item.py_name.replace('&', '&&'),
                            partial(self.context_menu_handler, action='edit_item_no_vl',
                                    index=index, ignore_vl=True))
                    self.context_menu.addAction(self.user_category_icon,
                            _('Add sub-category to %s')%item.py_name.replace('&', '&&'),
                            partial(self.context_menu_handler,
                                    action='add_subcategory', key=key))
                    self.context_menu.addAction(self.delete_icon,
                            _('Delete User category %s')%item.py_name.replace('&', '&&'),
                            partial(self.context_menu_handler,
                                    action='delete_user_category', key=key))
                    self.context_menu.addSeparator()
                # Add searches for temporary first letter nodes
                if self._model.collapse_model == 'first letter' and \
                        tag_item.temporary and not key.startswith('@'):
                    self.context_menu.addSeparator()
                    search_submenu = self.context_menu.addMenu(_('Search for'))
                    search_submenu.setIcon(QIcon.ic('search.png'))
                    search_submenu.addAction(self.search_icon,
                            '%s'%display_name(tag_item.tag),
                            partial(self.context_menu_handler, action='search',
                                    search_state=TAG_SEARCH_STATES['mark_plus'],
                                    index=index))
                    search_submenu.addAction(self.search_icon,
                            _('Everything but %s')%display_name(tag_item.tag),
                            partial(self.context_menu_handler, action='search',
                                    search_state=TAG_SEARCH_STATES['mark_minus'],
                                    index=index))
                # search by category. Some categories are not searchable, such
                # as search and news
                if item.tag.is_searchable:
                    if search_submenu is None:
                        search_submenu = self.context_menu.addMenu(_('Search for'))
                        search_submenu.setIcon(QIcon.ic('search.png'))
                        self.context_menu.addSeparator()
                    else:
                        search_submenu.addSeparator()
                    search_submenu.addAction(self.search_icon,
                            _('Books in category %s')%category,
                            partial(self.context_menu_handler,
                                    action='search_category',
                                    index=self._model.createIndex(item.row(), 0, item),
                                    search_state=TAG_SEARCH_STATES['mark_plus']))
                    search_submenu.addAction(self.search_icon,
                            _('Books not in category %s')%category,
                            partial(self.context_menu_handler,
                                    action='search_category',
                                    index=self._model.createIndex(item.row(), 0, item),
                                    search_state=TAG_SEARCH_STATES['mark_minus']))

                # Offer specific editors for tags/series/publishers/saved searches
                self.context_menu.addSeparator()
                if key in ['tags', 'publisher', 'series'] or (
                        fm['is_custom'] and fm['datatype'] != 'composite'):
                    if tag_item.type == TagTreeItem.CATEGORY and tag_item.temporary:
                        ac = self.context_menu.addAction(_('Manage %s')%category,
                            partial(self.context_menu_handler, action='open_editor',
                                    category=tag_item.name,
                                    key=key, is_first_letter=True))
                    else:
                        ac = self.context_menu.addAction(_('Manage %s')%category,
                            partial(self.context_menu_handler, action='open_editor',
                                    category=tag.original_name if tag else None,
                                    key=key))
                    ic = self.model().category_custom_icons.get(key)
                    if ic:
                        ac.setIcon(QIcon.ic(ic))
                    if fm['datatype'] == 'enumeration':
                        self.context_menu.addAction(_('Edit permissible values for %s')%category,
                            partial(self.context_menu_handler, action='edit_enum',
                                    key=key))
                elif key == 'authors':
                    if tag_item.type == TagTreeItem.CATEGORY:
                        if tag_item.temporary:
                            ac = self.context_menu.addAction(_('Manage %s')%category,
                                partial(self.context_menu_handler, action='edit_authors',
                                        index=tag_item.name, is_first_letter=True))
                        else:
                            ac = self.context_menu.addAction(_('Manage %s')%category,
                                partial(self.context_menu_handler, action='edit_authors'))
                    else:
                        ac = self.context_menu.addAction(_('Manage %s')%category,
                            partial(self.context_menu_handler, action='edit_authors',
                                    index=tag.id))
                    ic = self.model().category_custom_icons.get(key)
                    if ic:
                        ac.setIcon(QIcon.ic(ic))
                elif key == 'search':
                    self.context_menu.addAction(_('Manage Saved searches'),
                        partial(self.context_menu_handler, action='manage_searches',
                                category=tag.name if tag else None))
                elif key == 'formats' and tag is not None:
                    self.context_menu.addAction(_('Remove the {} format from selected books').format(tag.name),
                             partial(self.context_menu_handler, action='remove_format', key=tag.name))
                    self.context_menu.addSeparator()
                    self.context_menu.addAction(_('Add other application for %s files') % format(tag.name.upper()),
                             partial(self.context_menu_handler, action='add_open_with_apps', key=tag.name))
                    self.context_menu.addAction(_('Edit Open with applications for {} files').format(tag.name),
                             partial(self.context_menu_handler, action='edit_open_with_apps', key=tag.name))

                # Hide/Show/Restore categories
                self.context_menu.addSeparator()
                # Because of the strange way hierarchy works in user categories
                # where child nodes actually exist we must limit hiding to top-
                # level categories, which will hide that category and children
                if not key.startswith('@') or '.' not in key:
                    self.context_menu.addAction(_('Hide category %s') % category.replace('&', '&&'),
                        partial(self.context_menu_handler, action='hide',
                            category=key)).setIcon(QIcon.ic('minus.png'))
                add_show_hidden_categories()

                if tag is None:
                    cm = self.context_menu
                    cm.addSeparator()
                    acategory = category.replace('&', '&&')
                    sm = cm.addAction(_('Change {} category icon').format(acategory),
                                      partial(self.context_menu_handler, action='set_icon',
                                              key=key, category=category))
                    sm.setIcon(QIcon.ic('icon_choose.png'))
                    sm = cm.addAction(_('Restore {} category default icon').format(acategory),
                                      partial(self.context_menu_handler, action='clear_icon',
                                              key=key, category=category))
                    sm.setIcon(QIcon.ic('edit-clear.png'))
                    if key == 'search' and 'search' in self.db.new_api.pref('categories_using_hierarchy', ()):
                        sm = cm.addAction(_('Change Saved searches folder icon'),
                                          partial(self.context_menu_handler, action='set_icon',
                                                  key='search_folder:', category=_('Saved searches folder')))
                        sm.setIcon(QIcon.ic('icon_choose.png'))
                        sm = cm.addAction(_('Restore Saved searches folder default icon'),
                             partial(self.context_menu_handler, action='clear_icon',
                                     key='search_folder:', category=_('Saved searches folder')))
                        sm.setIcon(QIcon.ic('edit-clear.png'))

                # Always show the User categories editor
                self.context_menu.addSeparator()
                if key.startswith('@') and \
                        key[1:] in self.db.new_api.pref('user_categories', {}).keys():
                    self.context_menu.addAction(self.user_category_icon,
                            _('Manage User categories'),
                            partial(self.context_menu_handler, action='manage_categories',
                                    category=key[1:]))
                else:
                    self.context_menu.addAction(self.user_category_icon,
                            _('Manage User categories'),
                            partial(self.context_menu_handler, action='manage_categories',
                                    category=None))
        if self.hidden_categories:
            if not self.context_menu.isEmpty():
                self.context_menu.addSeparator()
            add_show_hidden_categories()

        # partitioning. If partitioning is active, provide a way to turn it on or
        # off for this category.
        if gprefs['tags_browser_partition_method'] != 'disable' and key is not None:
            m = self.context_menu
            p = self.db.prefs.get('tag_browser_dont_collapse', gprefs['tag_browser_dont_collapse'])
            if key in p:
                a = m.addAction(_('Sub-categorize {}').format(category),
                                partial(self.context_menu_handler, action='collapse_category',
                                        category=category, key=key, extra=p))
            else:
                a = m.addAction(_("Don't sub-categorize {}").format(category),
                                partial(self.context_menu_handler, action='dont_collapse_category',
                                        category=category, key=key, extra=p))
            a.setIcon(QIcon.ic('config.png'))
        # Set the partitioning scheme
        m = self.context_menu.addMenu(_('Change sub-categorization scheme'))
        m.setIcon(QIcon.ic('config.png'))
        da = m.addAction(_('Disable'),
            partial(self.context_menu_handler, action='categorization', category='disable'))
        fla = m.addAction(_('By first letter'),
            partial(self.context_menu_handler, action='categorization', category='first letter'))
        pa = m.addAction(_('Partition'),
            partial(self.context_menu_handler, action='categorization', category='partition'))
        if self.collapse_model == 'disable':
            da.setCheckable(True)
            da.setChecked(True)
        elif self.collapse_model == 'first letter':
            fla.setCheckable(True)
            fla.setChecked(True)
        else:
            pa.setCheckable(True)
            pa.setChecked(True)

        if config['sort_tags_by'] != "name":
            fla.setEnabled(False)
            m.hovered.connect(self.collapse_menu_hovered)
            fla.setToolTip(_('First letter is usable only when sorting by name'))
            # Apparently one cannot set a tooltip to empty, so use a star and
            # deal with it in the hover method
            da.setToolTip('*')
            pa.setToolTip('*')

        # Add expand menu items
        self.context_menu.addSeparator()
        m = self.context_menu.addMenu(_('Expand or collapse'))
        try:
            node_name = self._model.get_node(index).tag.name
        except AttributeError:
            pass
        else:
            if self.has_children(index) and not self.isExpanded(index):
                m.addAction(self.plus_icon,
                            _('Expand {0}').format(node_name), partial(self.expand, index))
            if self.has_unexpanded_children(index):
                m.addAction(self.plus_icon,
                            _('Expand {0} and its children').format(node_name),
                                            partial(self.expand_node_and_children, index))

        # Add menu items to collapse parent nodes
        idx = index
        paths = []
        while True:
            # First walk up the node tree getting the displayed names of
            # expanded parent nodes
            node = self._model.get_node(idx)
            if node.type == TagTreeItem.ROOT:
                break
            if self.has_children(idx) and self.isExpanded(idx):
                # leaf nodes don't have children so can't be expanded.
                # Also the leaf node might be collapsed
                paths.append((node.tag.name, idx))
            idx = self._model.parent(idx)
        for p in paths:
            # Now add the menu items
            m.addAction(self.minus_icon,
                        _("Collapse {0}").format(p[0]), partial(self.collapse_node, p[1]))
        m.addAction(self.minus_icon, _('Collapse all'), self.collapseAll)

        # Ask plugins if they have any actions to add to the context menu
        from calibre.gui2.ui import get_gui
        first = True
        for ac in get_gui().iactions.values():
            try:
                for context_action in ac.tag_browser_context_action(index):
                    if first:
                        self.context_menu.addSeparator()
                        first = False
                    self.context_menu.addAction(context_action)
            except Exception:
                import traceback
                traceback.print_exc()

        if not self.context_menu.isEmpty():
            self.context_menu.popup(self.mapToGlobal(point))
        return True

    def has_children(self, idx):
        return self.model().rowCount(idx) > 0

    def collapse_node_and_children(self, idx):
        self.collapse(idx)
        for r in range(self.model().rowCount(idx)):
            self.collapse_node_and_children(idx.child(r, 0))

    def collapse_node(self, idx):
        if not idx.isValid():
            return
        self.collapse_node_and_children(idx)
        self.setCurrentIndex(idx)
        self.scrollTo(idx)

    def expand_node_and_children(self, index):
        if not index.isValid():
            return
        self.expand(index)
        for r in range(self.model().rowCount(index)):
            self.expand_node_and_children(index.child(r, 0))

    def has_unexpanded_children(self, index):
        if not index.isValid():
            return False
        for r in range(self._model.rowCount(index)):
            dex = index.child(r, 0)
            if self._model.rowCount(dex) > 0:
                if not self.isExpanded(dex):
                    return True
                return self.has_unexpanded_children(dex)
        return False

    def collapse_menu_hovered(self, action):
        tip = action.toolTip()
        if tip == '*':
            tip = ''
        QToolTip.showText(QCursor.pos(), tip)

    def dragMoveEvent(self, event):
        QTreeView.dragMoveEvent(self, event)
        self.setDropIndicatorShown(False)
        index = self.indexAt(event.pos())
        if not index.isValid():
            return
        src_is_tb = event.mimeData().hasFormat('application/calibre+from_tag_browser')
        item = index.data(Qt.ItemDataRole.UserRole)
        if item.type == TagTreeItem.ROOT:
            return

        if src_is_tb:
            src_json = json_loads(bytes(event.mimeData().data('application/calibre+from_tag_browser')))
            if len(src_json) > 1:
                # Should never have multiple mimedata from the tag browser
                return
        if src_is_tb:
            src_md = src_json[0]
            src_item = self._model.get_node(self._model.index_for_path(src_md[5]))
            # Check if this is an intra-hierarchical-category drag/drop
            if (src_item.type == TagTreeItem.TAG and
                    src_item.tag.category == item.tag.category and
                    not item.temporary and
                    self._model.is_key_a_hierarchical_category(src_item.tag.category)):
                event.setDropAction(Qt.DropAction.MoveAction)
                self.setDropIndicatorShown(True)
                return
        # We aren't dropping an item on its own category. Check if the dest is
        # not a user category and can be dropped on. This covers drops from the
        # booklist. It is OK to drop onto virtual nodes
        if item.type == TagTreeItem.TAG and self._model.flags(index) & Qt.ItemFlag.ItemIsDropEnabled:
            event.setDropAction(Qt.DropAction.CopyAction)
            self.setDropIndicatorShown(not src_is_tb)
            return
        # Now see if we are on a user category and the source can be dropped there
        if item.type == TagTreeItem.CATEGORY and not item.is_gst:
            fm_dest = self.db.metadata_for_field(item.category_key)
            if fm_dest['kind'] == 'user':
                if src_is_tb:
                    # src_md and src_item are initialized above
                    if event.dropAction() == Qt.DropAction.MoveAction:
                        # can move only from user categories
                        if (src_md[0] == TagTreeItem.TAG and
                                 (not src_md[1].startswith('@') or src_md[2])):
                            return
                    # can't copy virtual nodes into a user category
                    if src_item.tag.is_editable:
                        self.setDropIndicatorShown(True)
                    return
                md = event.mimeData()
                # Check for drag to user category from the book list. Can handle
                # only non-multiple columns, except for some unknown reason authors
                if hasattr(md, 'column_name'):
                    fm_src = self.db.metadata_for_field(md.column_name)
                    if md.column_name in ['authors', 'publisher', 'series'] or \
                            (fm_src['is_custom'] and
                             ((fm_src['datatype'] in ['series', 'text', 'enumeration'] and
                                 not fm_src['is_multiple']) or
                              (fm_src['datatype'] == 'composite' and
                                  fm_src['display'].get('make_category', False)))):
                        self.setDropIndicatorShown(True)

    def clear(self):
        if self.model():
            self.model().clear_state()

    def is_visible(self, idx):
        item = idx.data(Qt.ItemDataRole.UserRole)
        if getattr(item, 'type', None) == TagTreeItem.TAG:
            idx = idx.parent()
        return self.isExpanded(idx)

    def recount_on_mark_change(self, *args):
        # Let other marked listeners run before we do the recount
        QTimer.singleShot(0, self.recount)

    def recount_with_position_based_index(self):
        self._model.use_position_based_index_on_next_recount = True
        self.recount()

    def recount(self, *args):
        '''
        Rebuild the category tree, expand any categories that were expanded,
        reset the search states, and reselect the current node.
        '''
        if self.disable_recounting or not self.pane_is_visible:
            return
        self.refresh_signal_processed = True
        ci = self.currentIndex()
        if not ci.isValid():
            ci = self.indexAt(QPoint(10, 10))
        use_pos = self._model.use_position_based_index_on_next_recount
        self._model.use_position_based_index_on_next_recount = False
        if use_pos:
            path = self._model.path_for_index(ci) if self.is_visible(ci) else None
        else:
            path = self._model.named_path_for_index(ci) if self.is_visible(ci) else None
        expanded_categories, state_map = self.get_state()
        self._model.rebuild_node_tree(state_map=state_map)
        self.blockSignals(True)
        for category in expanded_categories:
            idx = self._model.index_for_category(category)
            if idx is not None and idx.isValid():
                self.expand(idx)
        if path is not None:
            if use_pos:
                self.show_item_at_path(path)
            else:
                index = self._model.index_for_named_path(path)
                if index.isValid():
                    self.show_item_at_index(index)
        self.blockSignals(False)

    def show_item_at_path(self, path, box=False,
                          position=QAbstractItemView.ScrollHint.PositionAtCenter):
        '''
        Scroll the browser and open categories to show the item referenced by
        path. If possible, the item is placed in the center. If box=True, a
        box is drawn around the item.
        '''
        if path:
            self.show_item_at_index(self._model.index_for_path(path), box=box,
                                    position=position)

    def expand_parent(self, idx):
        # Needed otherwise Qt sometimes segfaults if the node is buried in a
        # collapsed, off screen hierarchy. To be safe, we expand from the
        # outermost in
        p = self._model.parent(idx)
        if p.isValid():
            self.expand_parent(p)
        self.expand(idx)

    def show_item_at_index(self, idx, box=False,
                           position=QAbstractItemView.ScrollHint.PositionAtCenter):
        if idx.isValid() and idx.data(Qt.ItemDataRole.UserRole) is not self._model.root_item:
            self.expand_parent(idx)
            self.setCurrentIndex(idx)
            self.scrollTo(idx, position)
            if box:
                self._model.set_boxed(idx)

    def item_expanded(self, idx):
        '''
        Called by the expanded signal
        '''
        self.setCurrentIndex(idx)

    # }}}



#!/usr/bin/env python


__license__   = 'GPL v3'
__copyright__ = '2013, Kovid Goyal <kovid at kovidgoyal.net>'
__docformat__ = 'restructuredtext en'

from calibre.ebooks import DRMError as _DRMError


class InvalidBook(ValueError):
    pass


class DRMError(_DRMError):

    def __init__(self):
        super().__init__(_('This file is locked with DRM. It cannot be edited.'))


class MalformedMarkup(ValueError):
    pass



# Copyright (C) 2006-2007 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#


from .element import Element
from .namespaces import TEXTNS
from .style import StyleElement

# Autogenerated


def A(**args):
    return Element(qname=(TEXTNS,'a'), **args)


def AlphabeticalIndex(**args):
    return Element(qname=(TEXTNS,'alphabetical-index'), **args)


def AlphabeticalIndexAutoMarkFile(**args):
    return Element(qname=(TEXTNS,'alphabetical-index-auto-mark-file'), **args)


def AlphabeticalIndexEntryTemplate(**args):
    return Element(qname=(TEXTNS,'alphabetical-index-entry-template'), **args)


def AlphabeticalIndexMark(**args):
    return Element(qname=(TEXTNS,'alphabetical-index-mark'), **args)


def AlphabeticalIndexMarkEnd(**args):
    return Element(qname=(TEXTNS,'alphabetical-index-mark-end'), **args)


def AlphabeticalIndexMarkStart(**args):
    return Element(qname=(TEXTNS,'alphabetical-index-mark-start'), **args)


def AlphabeticalIndexSource(**args):
    return Element(qname=(TEXTNS,'alphabetical-index-source'), **args)


def AuthorInitials(**args):
    return Element(qname=(TEXTNS,'author-initials'), **args)


def AuthorName(**args):
    return Element(qname=(TEXTNS,'author-name'), **args)


def Bibliography(**args):
    return Element(qname=(TEXTNS,'bibliography'), **args)


def BibliographyConfiguration(**args):
    return Element(qname=(TEXTNS,'bibliography-configuration'), **args)


def BibliographyEntryTemplate(**args):
    return Element(qname=(TEXTNS,'bibliography-entry-template'), **args)


def BibliographyMark(**args):
    return Element(qname=(TEXTNS,'bibliography-mark'), **args)


def BibliographySource(**args):
    return Element(qname=(TEXTNS,'bibliography-source'), **args)


def Bookmark(**args):
    return Element(qname=(TEXTNS,'bookmark'), **args)


def BookmarkEnd(**args):
    return Element(qname=(TEXTNS,'bookmark-end'), **args)


def BookmarkRef(**args):
    return Element(qname=(TEXTNS,'bookmark-ref'), **args)


def BookmarkStart(**args):
    return Element(qname=(TEXTNS,'bookmark-start'), **args)


def Change(**args):
    return Element(qname=(TEXTNS,'change'), **args)


def ChangeEnd(**args):
    return Element(qname=(TEXTNS,'change-end'), **args)


def ChangeStart(**args):
    return Element(qname=(TEXTNS,'change-start'), **args)


def ChangedRegion(**args):
    return Element(qname=(TEXTNS,'changed-region'), **args)


def Chapter(**args):
    return Element(qname=(TEXTNS,'chapter'), **args)


def CharacterCount(**args):
    return Element(qname=(TEXTNS,'character-count'), **args)


def ConditionalText(**args):
    return Element(qname=(TEXTNS,'conditional-text'), **args)


def CreationDate(**args):
    return Element(qname=(TEXTNS,'creation-date'), **args)


def CreationTime(**args):
    return Element(qname=(TEXTNS,'creation-time'), **args)


def Creator(**args):
    return Element(qname=(TEXTNS,'creator'), **args)


def DatabaseDisplay(**args):
    return Element(qname=(TEXTNS,'database-display'), **args)


def DatabaseName(**args):
    return Element(qname=(TEXTNS,'database-name'), **args)


def DatabaseNext(**args):
    return Element(qname=(TEXTNS,'database-next'), **args)


def DatabaseRowNumber(**args):
    return Element(qname=(TEXTNS,'database-row-number'), **args)


def DatabaseRowSelect(**args):
    return Element(qname=(TEXTNS,'database-row-select'), **args)


def Date(**args):
    return Element(qname=(TEXTNS,'date'), **args)


def DdeConnection(**args):
    return Element(qname=(TEXTNS,'dde-connection'), **args)


def DdeConnectionDecl(**args):
    return Element(qname=(TEXTNS,'dde-connection-decl'), **args)


def DdeConnectionDecls(**args):
    return Element(qname=(TEXTNS,'dde-connection-decls'), **args)


def Deletion(**args):
    return Element(qname=(TEXTNS,'deletion'), **args)


def Description(**args):
    return Element(qname=(TEXTNS,'description'), **args)


def EditingCycles(**args):
    return Element(qname=(TEXTNS,'editing-cycles'), **args)


def EditingDuration(**args):
    return Element(qname=(TEXTNS,'editing-duration'), **args)


def ExecuteMacro(**args):
    return Element(qname=(TEXTNS,'execute-macro'), **args)


def Expression(**args):
    return Element(qname=(TEXTNS,'expression'), **args)


def FileName(**args):
    return Element(qname=(TEXTNS,'file-name'), **args)


def FormatChange(**args):
    return Element(qname=(TEXTNS,'format-change'), **args)


def H(**args):
    return Element(qname=(TEXTNS, 'h'), **args)


def HiddenParagraph(**args):
    return Element(qname=(TEXTNS,'hidden-paragraph'), **args)


def HiddenText(**args):
    return Element(qname=(TEXTNS,'hidden-text'), **args)


def IllustrationIndex(**args):
    return Element(qname=(TEXTNS,'illustration-index'), **args)


def IllustrationIndexEntryTemplate(**args):
    return Element(qname=(TEXTNS,'illustration-index-entry-template'), **args)


def IllustrationIndexSource(**args):
    return Element(qname=(TEXTNS,'illustration-index-source'), **args)


def ImageCount(**args):
    return Element(qname=(TEXTNS,'image-count'), **args)


def IndexBody(**args):
    return Element(qname=(TEXTNS,'index-body'), **args)


def IndexEntryBibliography(**args):
    return Element(qname=(TEXTNS,'index-entry-bibliography'), **args)


def IndexEntryChapter(**args):
    return Element(qname=(TEXTNS,'index-entry-chapter'), **args)


def IndexEntryLinkEnd(**args):
    return Element(qname=(TEXTNS,'index-entry-link-end'), **args)


def IndexEntryLinkStart(**args):
    return Element(qname=(TEXTNS,'index-entry-link-start'), **args)


def IndexEntryPageNumber(**args):
    return Element(qname=(TEXTNS,'index-entry-page-number'), **args)


def IndexEntrySpan(**args):
    return Element(qname=(TEXTNS,'index-entry-span'), **args)


def IndexEntryTabStop(**args):
    return Element(qname=(TEXTNS,'index-entry-tab-stop'), **args)


def IndexEntryText(**args):
    return Element(qname=(TEXTNS,'index-entry-text'), **args)


def IndexSourceStyle(**args):
    return Element(qname=(TEXTNS,'index-source-style'), **args)


def IndexSourceStyles(**args):
    return Element(qname=(TEXTNS,'index-source-styles'), **args)


def IndexTitle(**args):
    return Element(qname=(TEXTNS,'index-title'), **args)


def IndexTitleTemplate(**args):
    return Element(qname=(TEXTNS,'index-title-template'), **args)


def InitialCreator(**args):
    return Element(qname=(TEXTNS,'initial-creator'), **args)


def Insertion(**args):
    return Element(qname=(TEXTNS,'insertion'), **args)


def Keywords(**args):
    return Element(qname=(TEXTNS,'keywords'), **args)


def LineBreak(**args):
    return Element(qname=(TEXTNS,'line-break'), **args)


def LinenumberingConfiguration(**args):
    return Element(qname=(TEXTNS,'linenumbering-configuration'), **args)


def LinenumberingSeparator(**args):
    return Element(qname=(TEXTNS,'linenumbering-separator'), **args)


def List(**args):
    return Element(qname=(TEXTNS,'list'), **args)


def ListHeader(**args):
    return Element(qname=(TEXTNS,'list-header'), **args)


def ListItem(**args):
    return Element(qname=(TEXTNS,'list-item'), **args)


def ListLevelStyleBullet(**args):
    return Element(qname=(TEXTNS,'list-level-style-bullet'), **args)


def ListLevelStyleImage(**args):
    return Element(qname=(TEXTNS,'list-level-style-image'), **args)


def ListLevelStyleNumber(**args):
    return Element(qname=(TEXTNS,'list-level-style-number'), **args)


def ListStyle(**args):
    return StyleElement(qname=(TEXTNS,'list-style'), **args)


def Measure(**args):
    return Element(qname=(TEXTNS,'measure'), **args)


def ModificationDate(**args):
    return Element(qname=(TEXTNS,'modification-date'), **args)


def ModificationTime(**args):
    return Element(qname=(TEXTNS,'modification-time'), **args)


def Note(**args):
    return Element(qname=(TEXTNS,'note'), **args)


def NoteBody(**args):
    return Element(qname=(TEXTNS,'note-body'), **args)


def NoteCitation(**args):
    return Element(qname=(TEXTNS,'note-citation'), **args)


def NoteContinuationNoticeBackward(**args):
    return Element(qname=(TEXTNS,'note-continuation-notice-backward'), **args)


def NoteContinuationNoticeForward(**args):
    return Element(qname=(TEXTNS,'note-continuation-notice-forward'), **args)


def NoteRef(**args):
    return Element(qname=(TEXTNS,'note-ref'), **args)


def NotesConfiguration(**args):
    return Element(qname=(TEXTNS,'notes-configuration'), **args)


def Number(**args):
    return Element(qname=(TEXTNS,'number'), **args)


def NumberedParagraph(**args):
    return Element(qname=(TEXTNS,'numbered-paragraph'), **args)


def ObjectCount(**args):
    return Element(qname=(TEXTNS,'object-count'), **args)


def ObjectIndex(**args):
    return Element(qname=(TEXTNS,'object-index'), **args)


def ObjectIndexEntryTemplate(**args):
    return Element(qname=(TEXTNS,'object-index-entry-template'), **args)


def ObjectIndexSource(**args):
    return Element(qname=(TEXTNS,'object-index-source'), **args)


def OutlineLevelStyle(**args):
    return Element(qname=(TEXTNS,'outline-level-style'), **args)


def OutlineStyle(**args):
    return Element(qname=(TEXTNS,'outline-style'), **args)


def P(**args):
    return Element(qname=(TEXTNS, 'p'), **args)


def Page(**args):
    return Element(qname=(TEXTNS,'page'), **args)


def PageContinuation(**args):
    return Element(qname=(TEXTNS,'page-continuation'), **args)


def PageCount(**args):
    return Element(qname=(TEXTNS,'page-count'), **args)


def PageNumber(**args):
    return Element(qname=(TEXTNS,'page-number'), **args)


def PageSequence(**args):
    return Element(qname=(TEXTNS,'page-sequence'), **args)


def PageVariableGet(**args):
    return Element(qname=(TEXTNS,'page-variable-get'), **args)


def PageVariableSet(**args):
    return Element(qname=(TEXTNS,'page-variable-set'), **args)


def ParagraphCount(**args):
    return Element(qname=(TEXTNS,'paragraph-count'), **args)


def Placeholder(**args):
    return Element(qname=(TEXTNS,'placeholder'), **args)


def PrintDate(**args):
    return Element(qname=(TEXTNS,'print-date'), **args)


def PrintTime(**args):
    return Element(qname=(TEXTNS,'print-time'), **args)


def PrintedBy(**args):
    return Element(qname=(TEXTNS,'printed-by'), **args)


def ReferenceMark(**args):
    return Element(qname=(TEXTNS,'reference-mark'), **args)


def ReferenceMarkEnd(**args):
    return Element(qname=(TEXTNS,'reference-mark-end'), **args)


def ReferenceMarkStart(**args):
    return Element(qname=(TEXTNS,'reference-mark-start'), **args)


def ReferenceRef(**args):
    return Element(qname=(TEXTNS,'reference-ref'), **args)


def Ruby(**args):
    return Element(qname=(TEXTNS,'ruby'), **args)


def RubyBase(**args):
    return Element(qname=(TEXTNS,'ruby-base'), **args)


def RubyText(**args):
    return Element(qname=(TEXTNS,'ruby-text'), **args)


def S(**args):
    return Element(qname=(TEXTNS,'s'), **args)


def Script(**args):
    return Element(qname=(TEXTNS,'script'), **args)


def Section(**args):
    return Element(qname=(TEXTNS,'section'), **args)


def SectionSource(**args):
    return Element(qname=(TEXTNS,'section-source'), **args)


def SenderCity(**args):
    return Element(qname=(TEXTNS,'sender-city'), **args)


def SenderCompany(**args):
    return Element(qname=(TEXTNS,'sender-company'), **args)


def SenderCountry(**args):
    return Element(qname=(TEXTNS,'sender-country'), **args)


def SenderEmail(**args):
    return Element(qname=(TEXTNS,'sender-email'), **args)


def SenderFax(**args):
    return Element(qname=(TEXTNS,'sender-fax'), **args)


def SenderFirstname(**args):
    return Element(qname=(TEXTNS,'sender-firstname'), **args)


def SenderInitials(**args):
    return Element(qname=(TEXTNS,'sender-initials'), **args)


def SenderLastname(**args):
    return Element(qname=(TEXTNS,'sender-lastname'), **args)


def SenderPhonePrivate(**args):
    return Element(qname=(TEXTNS,'sender-phone-private'), **args)


def SenderPhoneWork(**args):
    return Element(qname=(TEXTNS,'sender-phone-work'), **args)


def SenderPosition(**args):
    return Element(qname=(TEXTNS,'sender-position'), **args)


def SenderPostalCode(**args):
    return Element(qname=(TEXTNS,'sender-postal-code'), **args)


def SenderStateOrProvince(**args):
    return Element(qname=(TEXTNS,'sender-state-or-province'), **args)


def SenderStreet(**args):
    return Element(qname=(TEXTNS,'sender-street'), **args)


def SenderTitle(**args):
    return Element(qname=(TEXTNS,'sender-title'), **args)


def Sequence(**args):
    return Element(qname=(TEXTNS,'sequence'), **args)


def SequenceDecl(**args):
    return Element(qname=(TEXTNS,'sequence-decl'), **args)


def SequenceDecls(**args):
    return Element(qname=(TEXTNS,'sequence-decls'), **args)


def SequenceRef(**args):
    return Element(qname=(TEXTNS,'sequence-ref'), **args)


def SheetName(**args):
    return Element(qname=(TEXTNS,'sheet-name'), **args)


def SoftPageBreak(**args):
    return Element(qname=(TEXTNS,'soft-page-break'), **args)


def SortKey(**args):
    return Element(qname=(TEXTNS,'sort-key'), **args)


def Span(**args):
    return Element(qname=(TEXTNS,'span'), **args)


def Subject(**args):
    return Element(qname=(TEXTNS,'subject'), **args)


def Tab(**args):
    return Element(qname=(TEXTNS,'tab'), **args)


def TableCount(**args):
    return Element(qname=(TEXTNS,'table-count'), **args)


def TableFormula(**args):
    return Element(qname=(TEXTNS,'table-formula'), **args)


def TableIndex(**args):
    return Element(qname=(TEXTNS,'table-index'), **args)


def TableIndexEntryTemplate(**args):
    return Element(qname=(TEXTNS,'table-index-entry-template'), **args)


def TableIndexSource(**args):
    return Element(qname=(TEXTNS,'table-index-source'), **args)


def TableOfContent(**args):
    return Element(qname=(TEXTNS,'table-of-content'), **args)


def TableOfContentEntryTemplate(**args):
    return Element(qname=(TEXTNS,'table-of-content-entry-template'), **args)


def TableOfContentSource(**args):
    return Element(qname=(TEXTNS,'table-of-content-source'), **args)


def TemplateName(**args):
    return Element(qname=(TEXTNS,'template-name'), **args)


def TextInput(**args):
    return Element(qname=(TEXTNS,'text-input'), **args)


def Time(**args):
    return Element(qname=(TEXTNS,'time'), **args)


def Title(**args):
    return Element(qname=(TEXTNS,'title'), **args)


def TocMark(**args):
    return Element(qname=(TEXTNS,'toc-mark'), **args)


def TocMarkEnd(**args):
    return Element(qname=(TEXTNS,'toc-mark-end'), **args)


def TocMarkStart(**args):
    return Element(qname=(TEXTNS,'toc-mark-start'), **args)


def TrackedChanges(**args):
    return Element(qname=(TEXTNS,'tracked-changes'), **args)


def UserDefined(**args):
    return Element(qname=(TEXTNS,'user-defined'), **args)


def UserFieldDecl(**args):
    return Element(qname=(TEXTNS,'user-field-decl'), **args)


def UserFieldDecls(**args):
    return Element(qname=(TEXTNS,'user-field-decls'), **args)


def UserFieldGet(**args):
    return Element(qname=(TEXTNS,'user-field-get'), **args)


def UserFieldInput(**args):
    return Element(qname=(TEXTNS,'user-field-input'), **args)


def UserIndex(**args):
    return Element(qname=(TEXTNS,'user-index'), **args)


def UserIndexEntryTemplate(**args):
    return Element(qname=(TEXTNS,'user-index-entry-template'), **args)


def UserIndexMark(**args):
    return Element(qname=(TEXTNS,'user-index-mark'), **args)


def UserIndexMarkEnd(**args):
    return Element(qname=(TEXTNS,'user-index-mark-end'), **args)


def UserIndexMarkStart(**args):
    return Element(qname=(TEXTNS,'user-index-mark-start'), **args)


def UserIndexSource(**args):
    return Element(qname=(TEXTNS,'user-index-source'), **args)


def VariableDecl(**args):
    return Element(qname=(TEXTNS,'variable-decl'), **args)


def VariableDecls(**args):
    return Element(qname=(TEXTNS,'variable-decls'), **args)


def VariableGet(**args):
    return Element(qname=(TEXTNS,'variable-get'), **args)


def VariableInput(**args):
    return Element(qname=(TEXTNS,'variable-input'), **args)


def VariableSet(**args):
    return Element(qname=(TEXTNS,'variable-set'), **args)


def WordCount(**args):
    return Element(qname=(TEXTNS,'word-count'), **args)



__license__   = 'GPL v3'
__copyright__ = '2009, John Schember <john at nachtimwald.com>'
__docformat__ = 'restructuredtext en'

'''
Device driver for Barns and Nobel's Nook
'''

import io
import os

from calibre import fsync, prints
from calibre.constants import DEBUG
from calibre.devices.usbms.driver import USBMS
from calibre.utils.resources import get_image_path as I


class NOOK(USBMS):

    name           = 'Nook Device Interface'
    gui_name       = _('The Nook')
    description    = _('Communicate with the Nook e-book reader.')
    author         = 'John Schember'
    icon           = 'devices/nook.png'
    supported_platforms = ['windows', 'linux', 'osx']

    # Ordered list of supported formats
    FORMATS     = ['epub', 'pdb', 'pdf']

    VENDOR_ID   = [0x2080, 0x18d1]  # 0x18d1 is for softrooted nook
    PRODUCT_ID  = [0x001]
    BCD         = [0x322]

    VENDOR_NAME = 'B&N'
    WINDOWS_MAIN_MEM = 'NOOK'
    WINDOWS_CARD_A_MEM = 'NOOK'

    OSX_MAIN_MEM = 'B&N nook Media'
    OSX_CARD_A_MEM = OSX_MAIN_MEM

    MAIN_MEMORY_VOLUME_LABEL  = 'Nook Main Memory'
    STORAGE_CARD_VOLUME_LABEL = 'Nook Storage Card'

    EBOOK_DIR_MAIN = 'my documents'
    THUMBNAIL_HEIGHT = 144
    DELETE_EXTS = ['.jpg']
    SUPPORTS_SUB_DIRS = True

    def upload_cover(self, path, filename, metadata, filepath):
        from PIL import Image, ImageDraw

        coverdata = getattr(metadata, 'thumbnail', None)
        if coverdata and coverdata[2]:
            cover = Image.open(io.BytesIO(coverdata[2]))
        else:
            coverdata = open(I('library.png'), 'rb').read()

            cover = Image.new('RGB', (96, 144), 'black')
            im = Image.open(io.BytesIO(coverdata))
            im.thumbnail((96, 144), Image.Resampling.LANCZOS)

            x, y = im.size
            cover.paste(im, ((96-x)/2, (144-y)/2))

            draw = ImageDraw.Draw(cover)
            draw.text((1, 15), metadata.get('title', _('Unknown')).encode('ascii', 'ignore'))
            draw.text((1, 115), metadata.get('authors', _('Unknown')).encode('ascii', 'ignore'))

        data = io.BytesIO()
        cover.save(data, 'JPEG')
        coverdata = data.getvalue()

        with open('%s.jpg' % os.path.join(path, filename), 'wb') as coverfile:
            coverfile.write(coverdata)
            fsync(coverfile)

    def sanitize_path_components(self, components):
        return [x.replace('#', '_').replace('%', '_') for x in components]


class NOOK_COLOR(NOOK):
    name           = 'Nook Color Device Interface'
    description    = _('Communicate with the Nook Color, TSR, Glowlight and Tablet e-book readers.')

    PRODUCT_ID  = [
        0x002, 0x003, 0x004,
        0x005,  # Nook HD+
        0x007,  # Glowlight from 2013
        # 0xa,    # Glowlight from 2016 is MTP based device
        0xb,    # Glowlight from 2017
        0xc,    # Glowlight from 2019
        0xd,    # Glowlight from 2021
        0xe,    # Glowlight from 2022
        # 0xf,    # Glowlight from 2023 is MTP based device
    ]
    BCD         = [0x216, 0x9999, 0x409, 0x440]

    WINDOWS_MAIN_MEM = WINDOWS_CARD_A_MEM = ['EBOOK_DISK', 'NOOK_TABLET',
            'NOOK_SIMPLETOUCH', 'NOOK_GLOWLIGHT']
    EBOOK_DIR_MAIN = 'My Files'
    SCAN_FROM_ROOT = True
    NEWS_IN_FOLDER = False

    def upload_cover(self, path, filename, metadata, filepath):
        pass

    def post_open_callback(self):
        product_id = self.device_being_opened[1]
        if DEBUG:
            prints('Opened NOOK with product id:', product_id)
        if product_id >= 0xb:
            self.EBOOK_DIR_MAIN = 'NOOK/Books' if product_id >= 0xd else 'NOOK/My Files'
            if DEBUG:
                prints(f'Setting Nook upload directory to {self.EBOOK_DIR_MAIN}')
            try:
                os.makedirs(os.path.join(self._main_prefix, *self.EBOOK_DIR_MAIN.split('/')), exist_ok=True)
            except OSError:
                self.EBOOK_DIR_MAIN = 'NOOK'

    def get_carda_ebook_dir(self, for_upload=False):
        if for_upload:
            return self.EBOOK_DIR_MAIN
        return ''

    def create_upload_path(self, path, mdata, fname, create_dirs=True):
        is_news = mdata.tags and _('News') in mdata.tags
        subdir = 'Magazines' if is_news else 'Books'
        path = os.path.join(path, subdir)
        return USBMS.create_upload_path(self, path, mdata, fname,
                create_dirs=create_dirs)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>


import re
from collections import defaultdict, namedtuple
from functools import partial
from itertools import count
from operator import itemgetter

from css_parser.css import CSSRule, CSSStyleSheet, Property
from css_selectors import INAPPROPRIATE_PSEUDO_CLASSES, Select, SelectorError
from tinycss.fonts3 import parse_font_family, serialize_font_family

from calibre import as_unicode
from calibre.ebooks.css_transform_rules import all_properties
from calibre.ebooks.oeb.base import OEB_STYLES, SVG, XHTML, css_text
from calibre.ebooks.oeb.normalize_css import DEFAULTS, normalizers
from calibre.ebooks.oeb.stylizer import INHERITED, media_ok
from calibre.utils.resources import get_path as P
from polyglot.builtins import iteritems, itervalues

_html_css_stylesheet = None


def html_css_stylesheet(container):
    global _html_css_stylesheet
    if _html_css_stylesheet is None:
        data = P('templates/html.css', data=True).decode('utf-8')
        _html_css_stylesheet = container.parse_css(data, 'user-agent.css')
    return _html_css_stylesheet


def media_allowed(media):
    if not media or not media.mediaText:
        return True
    return media_ok(media.mediaText)


def iterrules(container, sheet_name, rules=None, media_rule_ok=media_allowed, rule_index_counter=None, rule_type=None, importing=None):
    ''' Iterate over all style rules in the specified sheet. Import and Media rules are
    automatically resolved. Yields (rule, sheet_name, rule_number).

    :param rules: List of CSSRules or a CSSStyleSheet instance or None in which case it is read from container using sheet_name
    :param sheet_name: The name of the sheet in the container (in case of inline style sheets, the name of the html file)
    :param media_rule_ok: A function to test if a @media rule is allowed
    :param rule_index_counter: A counter object, rule numbers will be calculated by incrementing the counter.
    :param rule_type: Only yield rules of this type, where type is a string type name, see css_parser.css.CSSRule for the names (
                    by default all rules are yielded)
    :return: (CSSRule object, the name of the sheet from which it comes, rule index - a monotonically increasing number)
    '''

    rule_index_counter = rule_index_counter or count()
    if importing is None:
        importing = set()
    importing.add(sheet_name)
    riter = partial(iterrules, container, rule_index_counter=rule_index_counter, media_rule_ok=media_rule_ok, rule_type=rule_type, importing=importing)
    if rules is None:
        rules = container.parsed(sheet_name)
    if rule_type is not None:
        rule_type = getattr(CSSRule, rule_type)

    for rule in rules:
        if rule.type == CSSRule.IMPORT_RULE:
            if media_rule_ok(rule.media):
                name = container.href_to_name(rule.href, sheet_name)
                if container.has_name(name):
                    if name in importing:
                        container.log.error(f'Recursive import of {name} from {sheet_name}, ignoring')
                    else:
                        csheet = container.parsed(name)
                        if isinstance(csheet, CSSStyleSheet):
                            yield from riter(name, rules=csheet)
        elif rule.type == CSSRule.MEDIA_RULE:
            if media_rule_ok(rule.media):
                yield from riter(sheet_name, rules=rule.cssRules)

        elif rule_type is None or rule.type == rule_type:
            num = next(rule_index_counter)
            yield rule, sheet_name, num

    importing.discard(sheet_name)


StyleDeclaration = namedtuple('StyleDeclaration', 'index declaration pseudo_element')
Specificity = namedtuple('Specificity', 'is_style num_id num_class num_elem rule_index')


def specificity(rule_index, selector, is_style=0):
    s = selector.specificity
    return Specificity(is_style, s[1], s[2], s[3], rule_index)


def iterdeclaration(decl):
    for p in all_properties(decl):
        n = normalizers.get(p.name)
        if n is None:
            yield p
        else:
            for k, v in iteritems(n(p.name, p.propertyValue)):
                yield Property(k, v, p.literalpriority)


class Values(tuple):

    ''' A tuple of `css_parser.css.Value ` (and its subclasses) objects. Also has a
    `sheet_name` attribute that is the canonical name relative to which URLs
    for this property should be resolved. '''

    def __new__(typ, pv, sheet_name=None, priority=''):
        ans = tuple.__new__(typ, pv)
        ans.sheet_name = sheet_name
        ans.is_important = priority == 'important'
        return ans

    @property
    def cssText(self):
        ' This will return either a string or a tuple of strings '
        if len(self) == 1:
            return css_text(self[0])
        return tuple(css_text(x) for x in self)


def normalize_style_declaration(decl, sheet_name):
    ans = {}
    for prop in iterdeclaration(decl):
        if prop.name == 'font-family':
            # Needed because of https://bitbucket.org/cthedot/cssutils/issues/66/incorrect-handling-of-spaces-in-font
            prop.propertyValue.cssText = serialize_font_family(parse_font_family(css_text(prop.propertyValue)))
        ans[prop.name] = Values(prop.propertyValue, sheet_name, prop.priority)
    return ans


def resolve_declarations(decls):
    property_names = set()
    for d in decls:
        property_names |= set(d.declaration)
    ans = {}
    for name in property_names:
        first_val = None
        for decl in decls:
            x = decl.declaration.get(name)
            if x is not None:
                if x.is_important:
                    first_val = x
                    break
                if first_val is None:
                    first_val = x
        ans[name] = first_val
    return ans


def resolve_pseudo_declarations(decls):
    groups = defaultdict(list)
    for d in decls:
        groups[d.pseudo_element].append(d)
    return {k:resolve_declarations(v) for k, v in iteritems(groups)}


def resolve_styles(container, name, select=None, sheet_callback=None):
    root = container.parsed(name)
    select = select or Select(root, ignore_inappropriate_pseudo_classes=True)
    style_map = defaultdict(list)
    pseudo_style_map = defaultdict(list)
    rule_index_counter = count()
    pseudo_pat = re.compile(':{1,2}(%s)' % ('|'.join(INAPPROPRIATE_PSEUDO_CLASSES)), re.I)

    def process_sheet(sheet, sheet_name):
        if sheet_callback is not None:
            sheet_callback(sheet, sheet_name)
        for rule, sheet_name, rule_index in iterrules(container, sheet_name, rules=sheet, rule_index_counter=rule_index_counter, rule_type='STYLE_RULE'):
            for selector in rule.selectorList:
                text = selector.selectorText
                try:
                    matches = tuple(select(text))
                except SelectorError as err:
                    container.log.error(f'Ignoring CSS rule with invalid selector: {text!r} ({as_unicode(err)})')
                    continue
                m = pseudo_pat.search(text)
                style = normalize_style_declaration(rule.style, sheet_name)
                if m is None:
                    for elem in matches:
                        style_map[elem].append(StyleDeclaration(specificity(rule_index, selector), style, None))
                else:
                    for elem in matches:
                        pseudo_style_map[elem].append(StyleDeclaration(specificity(rule_index, selector), style, m.group(1)))

    process_sheet(html_css_stylesheet(container), 'user-agent.css')

    for elem in root.iterdescendants(XHTML('style'), SVG('style'), XHTML('link')):
        if elem.tag.lower().endswith('style'):
            if not elem.text:
                continue
            sheet = container.parse_css(elem.text)
            sheet_name = name
        else:
            if (elem.get('type') or 'text/css').lower() not in OEB_STYLES or \
                    (elem.get('rel') or 'stylesheet').lower() != 'stylesheet' or \
                    not media_ok(elem.get('media')):
                continue
            href = elem.get('href')
            if not href:
                continue
            sheet_name = container.href_to_name(href, name)
            if not container.has_name(sheet_name):
                continue
            sheet = container.parsed(sheet_name)
            if not isinstance(sheet, CSSStyleSheet):
                continue
        process_sheet(sheet, sheet_name)

    for elem in root.xpath('//*[@style]'):
        text = elem.get('style')
        if text:
            style = container.parse_css(text, is_declaration=True)
            style_map[elem].append(StyleDeclaration(Specificity(1, 0, 0, 0, 0), normalize_style_declaration(style, name), None))

    for l in (style_map, pseudo_style_map):
        for x in itervalues(l):
            x.sort(key=itemgetter(0), reverse=True)

    style_map = {elem:resolve_declarations(x) for elem, x in iteritems(style_map)}
    pseudo_style_map = {elem:resolve_pseudo_declarations(x) for elem, x in iteritems(pseudo_style_map)}

    return partial(resolve_property, style_map), partial(resolve_pseudo_property, style_map, pseudo_style_map), select


_defvals = None


def defvals():
    global _defvals
    if _defvals is None:
        _defvals = {k:Values(Property(k, str(val)).propertyValue) for k, val in iteritems(DEFAULTS)}
    return _defvals


def resolve_property(style_map, elem, name):
    ''' Given a `style_map` previously generated by :func:`resolve_styles()` and
    a property `name`, returns the effective value of that property for the
    specified element. Handles inheritance and CSS cascading rules. Returns
    an instance of :class:`Values`. If the property was never set and
    is not a known property, then it will return None. '''

    inheritable = name in INHERITED
    q = elem
    while q is not None:
        s = style_map.get(q)
        if s is not None:
            val = s.get(name)
            if val is not None:
                return val
        q = q.getparent() if inheritable else None
    return defvals().get(name)


def resolve_pseudo_property(
    style_map, pseudo_style_map, elem, prop, name,
    abort_on_missing=False, check_if_pseudo_applies=False, check_ancestors=False
):
    if check_if_pseudo_applies:
        q = elem
        while q is not None:
            val = pseudo_style_map.get(q, {}).get(prop, {}).get(name)
            if val is not None:
                return True
            if not check_ancestors:
                break
            q = q.getparent()
        return False
    sub_map = pseudo_style_map.get(elem)
    if abort_on_missing and sub_map is None:
        return None
    if sub_map is not None:
        prop_map = sub_map.get(prop)
        if abort_on_missing and prop_map is None:
            return None
        if prop_map is not None:
            val = prop_map.get(name)
            if val is not None:
                return val
    if name in INHERITED:
        if check_ancestors:
            q = elem.getparent()
            while q is not None:
                val = pseudo_style_map.get(q, {}).get(prop, {}).get(name)
                if val is not None:
                    return val
                if not check_ancestors:
                    break
                q = q.getparent()
        return resolve_property(style_map, elem, name)
    return defvals().get(name)
